# Builtin
import logging
import time

# External
from Qt import QtWidgets, QtGui, QtCore

# Internal
from nxt_editor.dockwidgets.dock_widget_base import DockWidgetBase
from nxt import nxt_path
from nxt_editor.dockwidgets.layer_manager import LetterCheckboxDelegeate
import nxt_editor
from nxt_editor import colors

logger = logging.getLogger(nxt_editor.LOGGER_NAME)


class STARTS(object):
    SELECTION = 'Selection'
    FROM_SELECT = 'From selected'
    DESCENDANTS = 'Hierachy'  # TODO swap everywhere to desecendants.
    RUNNING = 'Running'


DEFAULT_STARTS = [
            STARTS.FROM_SELECT,
            STARTS.SELECTION,
            STARTS.DESCENDANTS]


class BuildView(DockWidgetBase):
    """Displays a combo box of start points as well as a list of the execution
    order implied by the selected start point. The combo box value can be
    edited to a node path to display execution order implied by that node path.
    Includes a button to execute the displayed order.
    """

    def __init__(self, parent=None):
        super(BuildView, self).__init__(title='Build', parent=parent)
        self.main_window = parent
        self.execute_actions = self.main_window.execute_actions
        self.execute_actions.pause_resume_exec_action.triggered.connect(self.pause_resume_pressed)
        self.stop_exec_action = self.execute_actions.stop_exec_action
        self.addAction(self.stop_exec_action)
        self.addAction(self.execute_actions.pause_resume_exec_action)
        self.main_widget = QtWidgets.QWidget()
        self.main_layout = QtWidgets.QVBoxLayout()
        self.main_widget.setLayout(self.main_layout)
        self.setWidget(self.main_widget)

        self.main_layout.addLayout(self.make_build_controls())
        self.build_table = BuildTable()
        self.main_layout.addWidget(self.build_table)

        self.pre_exec_start = ''
        self.in_running_mode = False

    def make_build_controls(self):
        """Assemble a Layout of build controls including a combo box and
        execution control buttons to go above the build table.

        :return: A Qt layout of buttons and a combobox
        :rtype: QtWidgets.QLayout
        """
        # Start combo box
        self.starts_combo = QtWidgets.QComboBox()
        self.starts_combo.setEditable(True)
        self.starts_combo.setInsertPolicy(QtWidgets.QComboBox.NoInsert)
        self.starts_combo.currentTextChanged.connect(self.start_text_changed)
        # Start action and button
        self.run_build_action = self.execute_actions.run_build_action
        self.run_build_action.triggered.connect(self.exec_build_pressed)
        self.restart_icon = QtGui.QIcon()
        rest_off_pix = QtGui.QPixmap(':icons/icons/reset.png')
        rest_hov_pix = QtGui.QPixmap(':icons/icons/reset_hover.png')
        self.restart_icon.addPixmap(rest_hov_pix, QtGui.QIcon.Active,
                                    QtGui.QIcon.Off)
        self.restart_icon.addPixmap(rest_off_pix, QtGui.QIcon.Normal,
                                    QtGui.QIcon.Off)
        self.restart_button = QtWidgets.QPushButton(self.restart_icon, '')
        self.restart_button.pressed.connect(self.run_build_action.trigger)
        self.restart_button.setEnabled(False)
        # Step button
        step_icon = self.execute_actions.step_build_action.icon()
        self.step_button = QtWidgets.QPushButton(step_icon, '')
        self.step_button.pressed.connect(self.step_build_pressed)
        # Pause Button
        self.pause_icon = QtGui.QIcon()
        pause_pixmap_on = QtGui.QPixmap(':icons/icons/pause.png')
        pause_pixmap_hov = QtGui.QPixmap(':icons/icons/pause_hover.png')
        self.pause_icon.addPixmap(pause_pixmap_on, QtGui.QIcon.Normal,
                             QtGui.QIcon.On)
        self.pause_icon.addPixmap(pause_pixmap_hov, QtGui.QIcon.Active,
                             QtGui.QIcon.Off)

        self.play_icon = QtGui.QIcon()
        play_pixmap_on = QtGui.QPixmap(':icons/icons/play.png')
        play_pixmap_hov = QtGui.QPixmap(':icons/icons/play_hover.png')
        self.play_icon.addPixmap(play_pixmap_on, QtGui.QIcon.Normal,
                             QtGui.QIcon.On)
        self.play_icon.addPixmap(play_pixmap_hov, QtGui.QIcon.Active,
                             QtGui.QIcon.Off)
        self.pause_resume_button = QtWidgets.QPushButton(self.play_icon, '')
        self.pause_resume_button.pressed.connect(self.pause_resume_pressed)
        # Stop button
        self.stop_button = QtWidgets.QPushButton(self.stop_exec_action.icon(),
                                                 '')
        self.stop_button.pressed.connect(self.stop_exec_action.trigger)

        self.controls_layout = QtWidgets.QHBoxLayout()
        self.controls_layout.addWidget(self.starts_combo, 1)
        self.controls_layout.addWidget(self.pause_resume_button, 0)
        self.controls_layout.addWidget(self.step_button, 0)
        self.controls_layout.addWidget(self.stop_button, 0)
        self.controls_layout.addWidget(self.restart_button, 0)
        return self.controls_layout

    def pause_resume_pressed(self):
        """When the pause resume button is pressed,
        pause or resume with context.
        """
        if self.stage_model.build_paused:
            self.stage_model.resume_build()
        elif self.in_running_mode:
            self.stage_model.pause_build()
        else:
            self.run_build_action.trigger()

    @property
    def visible_build(self):
        """If possible, retuns the list of nodes that are visible in the
        current build table.

        :return: list of node paths visible, if any.
        :rtype: list or None
        """
        if not self.build_model:
            return
        if not self.build_model.nodes:
            return
        return self.build_model.nodes

    def exec_build_pressed(self):
        """Exeucte current exec order(set in refresh_build_list)
        """
        vis_build = self.visible_build
        if not vis_build:
            return
        self.stage_model.execute_nodes(vis_build)

    def step_build_pressed(self):
        """When step build is presssed, either step the current build, or
        take first step into planned build.
        """
        vis_build = self.visible_build
        if not vis_build:
            return
        if not self.stage_model.is_build_setup():
            self.stage_model.setup_build(vis_build)
        self.stage_model.step_build()

    def set_stage_model(self, model):
        """Overload of dock widget base to disconnect previous model before
        changing and connect to new one after.
        """
        super(BuildView, self).set_stage_model(model)
        if not self.stage_model:
            return
        self.build_model = BuildModel(self.stage_model)
        self.build_table.setModel(self.build_model)
        self.main_widget.setEnabled(True)
        self.starts_combo.setEditText('')
        self.on_starts_changed(self.stage_model.get_start_nodes())
        self.on_executing_changed(self.stage_model.executing)
        self.on_model_focus_changed(self.stage_model.node_focus)
        return

    def set_stage_model_connections(self, model, connect):
        self.model_signal_connections = [
            (model.starts_changed, self.on_starts_changed),
            (model.comp_layer_changed, self.on_disp_layer_changed),
            (model.node_focus_changed, self.on_model_focus_changed),
            (model.executing_changed, self.on_executing_changed),
            (model.build_paused_changed, self.on_pause_change)
        ]
        super(BuildView, self).set_stage_model_connections(model, connect)

    def on_pause_change(self, paused):
        self.execute_actions.pause_resume_exec_action.setChecked(paused)
        if paused:
            icn = self.play_icon
        else:
            icn = self.pause_icon
        self.pause_resume_button.setIcon(icn)

    def on_executing_changed(self, executing):
        """Called when model's executing state is changed.
        Changes build view in and out of "running mode". Should disable
        and enable buttons as needed during execution.

        :param executing: Whether stage model is executing or not.
        :type executing: bool
        """
        self.stop_button.setEnabled(executing)
        self.stop_exec_action.setEnabled(executing)
        if executing == self.in_running_mode:
            return
        self.on_pause_change(not executing)
        self.in_running_mode = executing
        self.starts_combo.setEnabled(not self.in_running_mode)
        if self.in_running_mode:
            self.restart_button.setEnabled(True)
            self.pre_exec_start = self.starts_combo.currentText()
            self.starts_combo.setEditText(STARTS.RUNNING)
            self.stage_model.build_changed.connect(self.on_build_changed)
        else:
            self.restart_button.setEnabled(False)
            self.pause_resume_button.clearFocus()
            self.starts_combo.setEditText(self.pre_exec_start)

    def on_stage_model_destroyed(self):
        super(BuildView, self).on_stage_model_destroyed()
        self.build_table.setModel(None)
        self.main_widget.setEnabled(False)

    def on_disp_layer_changed(self, disp_layer):
        self.refresh_build_table()

    def start_text_changed(self, text):
        self.refresh_build_table()

    def on_build_changed(self, new_build):
        self.build_model.nodes = new_build

    def refresh_build_table(self):
        """Dump current build table is and rebuild it based on start combo.
        """
        # t0 = time.time()
        current_start = self.starts_combo.currentText()
        self.build_model.nodes = self.get_start_exec_order(current_start)
        # t1 = time.time()
        # logger.debug("Build Refresh took: " + str((t1-t0)))

    def get_start_exec_order(self, start):
        """Returns the execute order from given start.

        :param start: [description]
        :type start: [type]
        :return: [description]
        :rtype: [type]
        """
        # NOTE Duplicates some code from stage model, because it can't answer
        # the questions I needed to know. Implementations should merge someday
        if start == '':
            return []
        if start == nxt_path.WORLD:
            return []
        if start == STARTS.RUNNING:
            return self.stage_model.current_build_order
        if start not in DEFAULT_STARTS:
            # Must be an attempt at a literal node path
            if not self.stage_model.node_exists(start):
                return []
            # t2 = time.time()
            exec_order = self.stage_model.get_exec_order(start)
            # t3 = time.time()
            # logger.debug("Big output " + str(t3-t2))
            return exec_order
        sel_node_paths = self.stage_model.get_selected_nodes()
        if not sel_node_paths:
            return []
        if start == STARTS.SELECTION:
            return sel_node_paths
        if start == STARTS.FROM_SELECT:
            if len(sel_node_paths) != 1:
                return []
            start_path = sel_node_paths[0]
            return self.stage_model.get_exec_order(start_path)
        if start == STARTS.DESCENDANTS:
            if len(sel_node_paths) != 1:
                return []
            desc = self.stage_model.get_descendants(sel_node_paths[0])
            return sel_node_paths + desc

    def on_starts_changed(self, real_starts):
        """Called when a startpoint is added or removed to keep starts combo
        up to date.
        """
        prev_value = self.starts_combo.currentText()
        self.starts_combo.clear()
        if real_starts:
            self.starts_combo.addItems(real_starts)
        self.starts_combo.addItems(DEFAULT_STARTS)
        # Sometimes keep previous start point.
        if self.starts_combo.count() == len(DEFAULT_STARTS):
            self.starts_combo.setEditText('')
        elif prev_value != '':
            self.starts_combo.setEditText(prev_value)
        else:
            self.starts_combo.setEditText(real_starts[0])
        self.refresh_build_table()

    def on_model_focus_changed(self, new_focus):
        """When startpoint is concnered with selection, refresh table for
        focus changes.
        """
        if self.starts_combo.currentText() in DEFAULT_STARTS:
            self.refresh_build_table()


class BuildTable(QtWidgets.QTableView):
    """Displays a table of nodes for a build.
    The node highlighted in red will be the next run node if the graph is
    stepped or resumed.
    """
    def __init__(self):
        super(BuildTable, self).__init__()
        self.setSelectionMode(self.NoSelection)
        self.horizontalHeader().hide()
        self.verticalHeader().hide()
        self.break_delegate = LetterCheckboxDelegeate('B')
        self.setItemDelegateForColumn(BuildModel.BREAK_COLUMN,
                                      self.break_delegate)
        self.skip_delegate = LetterCheckboxDelegeate('S')
        self.setItemDelegateForColumn(BuildModel.SKIP_COLUMN,
                                      self.skip_delegate)
        self.clicked.connect(self.on_row_clicked)
        # TODO context menu and shift-click for +descendents.

    def on_row_clicked(self, clicked_idx):
        """When a row in the table is clicked, select and frame.
        """
        if clicked_idx.column() != BuildModel.PATH_COLUMN:
            return
        path_clicked = self.model()._nodes[clicked_idx.row()]
        self.model().stage_model.select_and_frame(path_clicked)

    def setModel(self, model):
        """Sets up headers.
        """
        super(BuildTable, self).setModel(model)
        header = self.horizontalHeader()
        header.setStretchLastSection(False)
        header.setDefaultSectionSize(28)
        header.setSectionResizeMode(header.Fixed)
        if header.count():
            column = BuildModel.PATH_COLUMN
            header.setSectionResizeMode(column, QtWidgets.QHeaderView.Stretch)
            self.hideColumn(BuildModel.START_COLUMN)
            self.hideColumn(BuildModel.NEXT_RUN_COLUMN)
        if not model:
            return
        model.stage_model.build_idx_changed.connect(self.on_build_idx_changed)

    def on_build_idx_changed(self, build_idx):
        if self.model().stage_model.can_build_run():
            model_index = self.model().index(build_idx, 0)
        else:
            model_index = self.model().index(0, 0)
        self.scrollTo(model_index, self.ScrollHint.PositionAtCenter)


class BuildModel(QtCore.QAbstractTableModel):
    """A model of a series of nodes that reflects execution information
    including break status, start status, and which node is next to be run.
    """
    SKIP_COLUMN = 0
    BREAK_COLUMN = 1
    START_COLUMN = 2
    PATH_COLUMN = 3
    NEXT_RUN_COLUMN = 4

    def __init__(self, stage_model):
        super(BuildModel, self).__init__()
        """self._nodes is the execute order this build model with reflect
        answers about.
        """
        self.headers = ['Skip', 'Break', 'Start', 'Path', 'Next']
        self.stage_model = stage_model
        self._nodes = []
        self.stage_model.skips_changed.connect(self.on_skips_changed)
        self.stage_model.breaks_changed.connect(self.on_breaks_changed)
        self.stage_model.build_idx_changed.connect(self.on_build_idx_changed)

    @property
    def nodes(self):
        return self._nodes

    @nodes.setter
    def nodes(self, val):
        """Allows a direct reset of the model into a new execution order.
        """
        if val == self.nodes:
            return
        self.beginResetModel()
        self._nodes = val
        self.endResetModel()

    def on_skips_changed(self, new_skips):
        last_row = len(self.nodes) - 1
        self.dataChanged.emit(self.index(0, self.SKIP_COLUMN),
                              self.index(last_row, self.SKIP_COLUMN))

    def on_breaks_changed(self, new_breaks):
        last_row = len(self.nodes) - 1
        self.dataChanged.emit(self.index(0, self.BREAK_COLUMN),
                              self.index(last_row, self.BREAK_COLUMN))

    def on_build_idx_changed(self, new_idx):
        if not self.nodes:
            return
        last_row = len(self.nodes) - 1
        self.dataChanged.emit(self.index(0, self.PATH_COLUMN),
                              self.index(last_row, self.NEXT_RUN_COLUMN))

    def headerData(self, section, orientation, role):
        if orientation == QtCore.Qt.Horizontal:
            if role == QtCore.Qt.DisplayRole:
                return self.headers[section]

    def rowCount(self, index):
        if index.column() > 0 or not self.nodes:
            return 0
        return len(self.nodes)

    def columnCount(self, index):
        return len(self.headers)

    def data(self, index, role=None):
        row = index.row()
        column = index.column()
        idx_path = self.nodes[row]
        if role == QtCore.Qt.DisplayRole:
            if column == self.PATH_COLUMN:
                return idx_path
        next_run = False
        if self.stage_model.is_build_setup():
            next_run = row == self.stage_model.last_built_idx
        elif row == 0:
            next_run = True
        is_start = self.stage_model.get_is_node_start(idx_path)
        is_break = self.stage_model.get_is_node_breakpoint(idx_path)
        is_skip = self.stage_model.is_node_skippoint(idx_path)
        if role == QtCore.Qt.CheckStateRole:
            if column == self.START_COLUMN:
                return QtCore.Qt.Checked if is_start else QtCore.Qt.Unchecked
            if column == self.BREAK_COLUMN:
                return QtCore.Qt.Checked if is_break else QtCore.Qt.Unchecked
            if column == self.SKIP_COLUMN:
                return QtCore.Qt.Checked if is_skip else QtCore.Qt.Unchecked
            if column == self.NEXT_RUN_COLUMN:
                return QtCore.Qt.Checked if next_run else QtCore.Qt.Unchecked
        if role == QtCore.Qt.BackgroundRole:
            if column == self.BREAK_COLUMN and is_break:
                return colors.BREAK_COLOR
            if column == self.SKIP_COLUMN and is_skip:
                return colors.SKIP_COLOR
            if column == self.START_COLUMN and is_start:
                return colors.START_COLOR
            if column == self.PATH_COLUMN and next_run:
                return QtGui.QBrush(QtCore.Qt.red)
        if role == QtCore.Qt.ForegroundRole:
            if column == self.PATH_COLUMN:
                return QtGui.QBrush(QtCore.Qt.white)

    def setData(self, index, value, role=QtCore.Qt.EditRole):
        column = index.column()
        if role != QtCore.Qt.CheckStateRole:
            return False
        path = self.nodes[index.row()]
        if column == self.BREAK_COLUMN:
            current_break = self.stage_model.get_is_node_breakpoint(path)
            self.stage_model.set_breakpoints([path], not current_break)
            return True
        if column == self.SKIP_COLUMN:
            modifiers = QtWidgets.QApplication.keyboardModifiers()
            if modifiers == QtCore.Qt.ShiftModifier:
                self.stage_model.toggle_descendant_skips([path])
            else:
                self.stage_model.toggle_skippoints([path])
            return True
        return False



# Builtin
import os
import json

# Internal
from nxt.constants import USER_DIR

EDITOR_DIR = os.path.dirname(__file__)
version_file = os.path.join(EDITOR_DIR, 'version.json')


class EDITOR_VERSION(object):
    with open(version_file, 'r') as f:
        version_data = json.load(f)
    api_v_data = version_data['EDITOR']
    MAJOR = api_v_data['MAJOR']
    MINOR = api_v_data['MINOR']
    PATCH = api_v_data['PATCH']
    VERSION_TUPLE = (MAJOR, MINOR, PATCH)
    VERSION_STR = '.'.join(str(v) for v in VERSION_TUPLE)
    VERSION = VERSION_STR


class FONTS(object):
    DEFAULT_FAMILY = 'RobotoMono-Regular'
    DEFAULT_SIZE = 10


_pref_dir_name = str(EDITOR_VERSION.MAJOR)
PREF_DIR = os.path.join(USER_DIR, 'prefs', _pref_dir_name)

NXT_WEBSITE = 'https://nxt-dev.github.io/'



# Built-in
import os
import sys
import logging
import subprocess
import traceback
from collections import OrderedDict
import webbrowser
from functools import partial
import time

# External
from Qt import QtWidgets
from Qt import QtGui
from Qt import QtCore

# Internal
import nxt_editor
from nxt_editor import user_dir
from nxt.session import Session
from nxt_editor.constants import EDITOR_VERSION, FONTS
from nxt_editor.stage_view import StageView
from nxt_editor.stage_model import StageModel
from nxt_editor.dockwidgets import (DockWidgetBase, CodeEditor, PropertyEditor,
                                    HotkeyEditor, LayerManager, OutputLog,
                                    HistoryView, WidgetBuilder, BuildView,
                                    FindRepDockWidget)
from nxt_editor.dockwidgets.output_log import (FileTailingThread,
                                               QtLogStreamHandler)
from nxt_editor.dockwidgets.code_editor import NxtCodeEditor
from nxt import nxt_log, nxt_io, nxt_layer
from nxt_editor.dialogs import (NxtFileDialog, NxtWarningDialog,
                                UnsavedLayersDialogue, UnsavedChangesMessage)
from nxt_editor import actions, LoggingSignaler
from nxt.constants import (API_VERSION, GRAPH_VERSION, USER_PLUGIN_DIR,
                           NXT_DCC_ENV_VAR, is_standalone)
from nxt.remote.client import NxtClient
import nxt.remote.contexts
from nxt_editor import qresources


logger = logging.getLogger(nxt_editor.LOGGER_NAME)


class MainWindow(QtWidgets.QMainWindow):

    """The main window of the nxt UI. Includes the menu bar, tool bar, and dock widgets."""

    tab_changed = QtCore.Signal()
    close_signal = QtCore.Signal()
    new_log_signal = QtCore.Signal(logging.LogRecord)
    font_size_changed = QtCore.Signal(int)

    def __init__(self, filepath=None, parent=None, start_rpc=True):
        """Create NXT window.

        :param parent: parent to attach this UI to.
        :type parent: QtWidgets.QtWidgets.QWidget
        """
        self.in_startup = True
        pixmap = QtGui.QPixmap(':icons/icons/nxt.svg')
        self.splash_screen = QtWidgets.QSplashScreen(pixmap)
        self.splash_screen.show()
        self.splash_screen.showMessage('Starting nxt...',
                                       QtCore.Qt.AlignCenter, QtCore.Qt.white)
        QtWidgets.QApplication.processEvents()
        super(MainWindow, self).__init__(parent=parent)
        self.new_log_signal.connect(self.handle_remote_log)
        old_cwd = os.getcwd()
        ui_dir = os.path.dirname(__file__)
        os.chdir(ui_dir)
        # Test to see if we're launching from a git branch, if so the title
        # bar will be updated for easy reference.

        # Used to hide the stderr from the user as it doesn't matter
        f = open(nxt_io.generate_temp_file('NxtGitErr'))
        try:
            git_out = subprocess.check_output(["git", "branch"],
                                              stderr=f).decode("utf8")
            cur = next(line for line in git_out.split("\n")
                       if line.startswith("*"))
            current_branch = cur.strip("*").strip()
        except:  # Broad because Maya
            # Failed to run git branch, attempting fallback method
            try:
                with open('../../.git/HEAD') as f:
                    head = f.read()
                _, __, current_branch = head.rpartition('/')
            except:
                # Could not determine git branch, must be pip package.
                current_branch = ''
        finally:
            f.close()
        os.chdir(old_cwd)
        if is_standalone():
            context = 'standalone'
        else:
            context = os.environ.get(NXT_DCC_ENV_VAR) or ''
        self.host_app = context
        self.setWindowTitle("nxt {} - Editor v{} | Graph v{} | API v{} "
                            "(Python {}) {}".format(self.host_app,
                                                    EDITOR_VERSION.VERSION_STR,
                                                    GRAPH_VERSION.VERSION_STR,
                                                    API_VERSION.VERSION_STR,
                                                    '.'.join([str(n) for n in sys.version_info[:3]]),
                                                    current_branch))
        self.setObjectName('Main Window')
        self.zoom_keys = QtGui.QKeySequence(QtCore.Qt.Key_Alt)
        self.zoom_keys_down = False
        self._held_keys = []
        self._closing = False
        self.last_focused_start = 0  # Start point focus tracker
        # FIXME: Fix with MV signal
        self.last_focused_tab = -1  # Tab tracker for upating the comp layer
        # set app icon
        self.app_icon = QtGui.QIcon(pixmap)
        self.setWindowIcon(self.app_icon)

        # set style sheet
        style_file = QtCore.QFile(':styles/styles/dark/dark.qss')
        style_file.open(QtCore.QFile.ReadOnly)
        self.stylesheet = str(style_file.readAll())
        self.setStyleSheet(self.stylesheet)

        # fonts
        font_db = QtGui.QFontDatabase()
        font_db.addApplicationFont(":fonts/fonts/RobotoMono/RobotoMono-Regular.ttf")
        font_db.addApplicationFont(":fonts/fonts/Roboto/Roboto-Regular.ttf")

        # nxt object in charge of loaded graphs
        self.nxt = Session()

        # APPLICATION WIDE ACTIONS
        # TODO: All the actions should be connected to functions in nxt not
        #  view
        self.splash_screen.showMessage('Setting up hotkeys...',
                                       QtCore.Qt.AlignCenter, QtCore.Qt.white)
        self.app_actions = actions.AppActions(self)
        self.addActions(self.app_actions.actions())
        # NODE ACTIONS
        self.node_actions = actions.NodeActions(self)
        # PROPERTY ACTIONS
        self.property_manager_actions = actions.PropertyEditorActions(self)
        # NODE COMMENT ACTIONS
        self.node_comment_actions = actions.NodeCommentActions(self)
        # LAYER ACTIONS
        self.layer_actions = actions.LayerActions(self)
        # ALIGNMENT ACTIONS
        self.alignment_actions = actions.AlignmentActions(self)
        # DISPLAY ACTIONS
        self.display_actions = actions.DisplayActions(self)
        # VIEW ACTIONS
        self.view_actions = actions.StageViewActions(self)
        # EXEC ACTIONS
        self.execute_actions = actions.ExecuteActions(self)
        self.addAction(self.execute_actions.stop_exec_action)
        # CODE EDITOR ACTIONS
        self.code_editor_actions = actions.CodeEditorActions(self)
        # TOOL BARS
        self.authoring_toolbar = NodeAuthoringToolBar(self,
                                                      self.font_size_changed)
        self.addToolBar(self.authoring_toolbar)
        self.execute_toolbar = ExecuteToolBar(self, self.font_size_changed)
        self.addToolBar(self.execute_toolbar)
        self.display_toolbar = DisplayToolBar(self, self.font_size_changed)
        self.addToolBar(self.display_toolbar)
        self.align_distribute_toolbar = AlignDistributeToolBar(self,
                                                               self.font_size_changed)
        self.addToolBar(self.align_distribute_toolbar)
        # TABS WIDGET
        self.open_files_tab_widget = OpenFilesTabWidget(parent=self)
        self.open_files = {}  # TODO: Doesn't this duplicate what Nxt does?
        self.previous_view = None
        # graph tabs
        self.open_files_tab_widget.currentChanged.connect(self.on_tab_change)
        self.setCentralWidget(self.open_files_tab_widget)
        self.splash_screen.showMessage('Setting up dockwidgets...',
                                       QtCore.Qt.AlignCenter, QtCore.Qt.white)
        # Dock Widgets
        # hotkey editor
        self.hotkey_editor = HotkeyEditor(parent=self)
        self.hotkey_editor.hide()

        # property editor
        self.property_editor = PropertyEditor(parent=self)
        self.addDockWidget(QtCore.Qt.RightDockWidgetArea, self.property_editor)

        # code editor
        self.code_editor = CodeEditor(parent=self)
        self.code_editor.editor.viewport().installEventFilter(self)
        self.addDockWidget(QtCore.Qt.RightDockWidgetArea, self.code_editor)

        # Find and Replace
        self.find_rep = FindRepDockWidget(parent=self)
        self.addDockWidget(QtCore.Qt.BottomDockWidgetArea, self.find_rep)
        self.find_rep.hide()
        # layer manager
        self.layer_manager = LayerManager(parent=self)
        self.addDockWidget(QtCore.Qt.LeftDockWidgetArea, self.layer_manager)

        # history view
        self.history_view = HistoryView(parent=self)
        self.addDockWidget(QtCore.Qt.LeftDockWidgetArea, self.history_view)

        # build View
        self.build_view = BuildView(parent=self)
        self.addDockWidget(QtCore.Qt.LeftDockWidgetArea, self.build_view)

        # output log
        self.output_log = OutputLog(parent=self)
        self.output_log.hide()
        self.addDockWidget(QtCore.Qt.BottomDockWidgetArea, self.output_log)

        # workflow tools
        self.workflow_tools = WidgetBuilder(parent=self)
        self.addDockWidget(QtCore.Qt.LeftDockWidgetArea, self.workflow_tools)

        self.setCorner(QtCore.Qt.BottomRightCorner,
                       QtCore.Qt.RightDockWidgetArea)
        self.setCorner(QtCore.Qt.BottomLeftCorner,
                       QtCore.Qt.LeftDockWidgetArea)
        self.setTabPosition(QtCore.Qt.AllDockWidgetAreas,
                            QtWidgets.QTabWidget.North)

        # status bar
        self.status_bar = QtWidgets.QStatusBar()
        self.status_bar.setSizeGripEnabled(False)
        self.status_bar.setContentsMargins(4, 4, 4, 4)
        self.status_bar.setStyleSheet('color: lightGrey; background-color: #232323; border: 4px solid #3E3E3E')
        self.setStatusBar(self.status_bar)

        self.log_button = QtWidgets.QPushButton("Show Log")
        self.log_button.setMinimumWidth(75)
        self.log_button.setStyleSheet(self.stylesheet)
        self.output_log.visibilityChanged.connect(self.refresh_log_button)
        self.log_button.clicked.connect(self.log_button_clicked)
        self.status_bar.addPermanentWidget(self.log_button)
        self.refresh_log_button()

        self.logger = logging.getLogger('nxt')
        self.logger.addHandler(StatusBarHandler(self.status_bar))

        self.state_last_hidden = None
        # TODO set and load default geometry
        # TODO determine and create sensible default position and size for the window, perhaps 80% of available screen?
        # print QDesktopWidget.availableGeometry(self)
        self.resize(1600, 800)
        self.resizeDocks([self.property_editor, self.code_editor], [400, 300], QtCore.Qt.Vertical)

        if filepath:
            self.load_file(filepath=filepath)
        else:
            self.new_tab()
        # menu bar
        # TODO: Depends on dock widgets this should change
        self.menu_bar = MenuBar(self)
        self.setMenuBar(self.menu_bar)
        self.menuBar().setNativeMenuBar(False)
        self.display_actions.resolve_action.setChecked(True)
        # Rpc startup
        self.rpc_log_tail = None
        if start_rpc:
            self.startup_rpc_server(join=False)
        self.splash_screen.showMessage('Restoring \nfont \nsize...',
                                       QtCore.Qt.AlignCenter,
                                       QtCore.Qt.white)
        self.restore_font_size()
        # Should this be a signal? Like Startup done, now you can refresh?
        self.splash_screen.finish(self)
        self.in_startup = False
        t = QtCore.QTimer()
        t.setInterval(256)

        def failure_check():
            if self.view:
                self.view.failure_check()
            t.stop()
        t.timeout.connect(failure_check)
        t.start()

        app = QtWidgets.QApplication.instance()
        app.aboutToQuit.connect(self.shutdown_rpc_server)

    # RPC
    def startup_rpc_server(self, join=True):
        t = StartRPCThread(self)
        t.start()
        if join:
            t.wait()
        else:
            txt = 'Waiting on rpc server...'
            txt_len = len(txt)
            self.count = 0

            def tick():
                self.splash_screen.showMessage(txt[:self.count % -txt_len],
                                               QtCore.Qt.AlignCenter,
                                               QtCore.Qt.white)
                self.count += 1
            timer = QtCore.QTimer()
            timer.setInterval(100)
            timer.timeout.connect(tick)
            t.finished.connect(timer.stop)
            timer.start()
            while not t.isFinished():
                QtWidgets.QApplication.processEvents()

    @staticmethod
    def handle_remote_log(record):
        logger.handle(record)

    def shutdown_rpc_server(self):
        if self.model:
            self.model.processing.emit(True)
        self.safe_stop_rpc_tailing()
        self.nxt.shutdown_rpc_server()
        if self.model:
            self.model.processing.emit(False)
        if not self.rpc_log_tail:
            return
        wait_started = time.time()
        while not self.rpc_log_tail.isFinished():
            QtWidgets.QApplication.processEvents()
            if time.time() - wait_started > 5:
                logger.error('Failed to stop rpc log tail!')
                return
        self.rpc_log_tail = None

    def safe_stop_rpc_tailing(self):
        if not self.rpc_log_tail:
            return
        self.handle_rpc_tailing_signals(False)
        self.rpc_log_tail.requestInterruption()

    def handle_rpc_tailing_signals(self, state):
        if not self.rpc_log_tail:
            return
        raw_write_func = self.output_log._write_raw_output
        rich_write_func = self.output_log.write_rich_output
        if state:
            self.rpc_log_tail.new_text.connect(raw_write_func)
            self.rpc_log_tail.new_text.connect(rich_write_func)
        else:
            self.rpc_log_tail.new_text.disconnect(raw_write_func)
            self.rpc_log_tail.new_text.disconnect(rich_write_func)

    def event(self, event):
        if event.type() == QtCore.QEvent.WindowDeactivate:
            self._held_keys = []
            self.zoom_keys_down = False
        return super(MainWindow, self).event(event)

    @staticmethod
    def set_waiting_cursor(state=True):
        if state:
            QtWidgets.QApplication.setOverrideCursor(QtCore.Qt.WaitCursor)
        else:
            QtWidgets.QApplication.restoreOverrideCursor()

    def increase_font_size(self):
        self._change_font_size(1)

    def decrease_font_size(self):
        self._change_font_size(-1)

    def _change_font_size(self, delta, absolute=False, save=True):
        app = QtWidgets.QApplication.instance()
        if absolute:
            font_size = delta
        else:
            font_size = app.font().pointSize() + delta
            self.font_size_changed.emit(delta)
        if save:
            user_dir.user_prefs[user_dir.USER_PREF.FONT_SIZE] = font_size
        font = QtGui.QFont(FONTS.DEFAULT_FAMILY, font_size)
        app.setFont(font)

        widgets_with_fonts = ["QMenuBar", "QTabWidget", "QMenu", "QTableView",
                              "QLineEdit", "QComboBox", "QLabel",
                              "QPushButton", "QTextEdit", "QWidget",
                              "QListWidget", "QTabelWidget", "QTreeWidget",
                              "QSpinBox", "QDoubleSpinBox", "QCheckBox"]
        for widget in widgets_with_fonts:
            app.setFont(font, widget)

        new_cb_stylesheet = '''
QCheckBox::indicator {
    margin-left: 4px;
    width: %spx;
    height: %spx;
}
                    ''' % (font_size * 1.5, font_size * 1.5)
        # List all checkboxes and update their style sheet.
        for widget in QtWidgets.QApplication.allWidgets():
            if isinstance(widget, QtWidgets.QCheckBox):
                widget.setStyleSheet(new_cb_stylesheet)

    def restore_font_size(self):
        user_pref_size = user_dir.user_prefs.get(user_dir.USER_PREF.FONT_SIZE,
                                                 FONTS.DEFAULT_SIZE)
        if user_pref_size > FONTS.DEFAULT_SIZE:
            self.font_size_changed.emit(user_pref_size - FONTS.DEFAULT_SIZE)
        elif user_pref_size < FONTS.DEFAULT_SIZE:
            self.font_size_changed.emit(FONTS.DEFAULT_SIZE - user_pref_size)
        self._change_font_size(user_pref_size, absolute=True, save=False)

    def reset_font_size(self):
        self._change_font_size(FONTS.DEFAULT_SIZE, absolute=True, save=False)
        user_dir.user_prefs.pop(user_dir.USER_PREF.FONT_SIZE)
        self.font_size_changed.emit(0)


    @staticmethod
    def create_remote_context(place_holder_text='',
                              interpreter_exe=sys.executable,
                              context_graph=None, exe_script_args=()):
        cur_context = nxt.remote.contexts.get_current_context_exe_name()
        pop_up = QtWidgets.QDialog()
        pop_up.setWindowTitle('Create context for "{}"'.format(cur_context))
        v_layout = QtWidgets.QVBoxLayout()
        pop_up.setLayout(v_layout)
        label = QtWidgets.QPlainTextEdit()
        info = ('Create remote context for your host '
                'Python interpreter/DCC\n'
                'Type your desired name in the box below '
                'and click create.'.format(cur_context))
        label.setPlainText(info)
        label.setReadOnly(True)
        font_metric = QtGui.QFontMetrics(label.document().defaultFont())
        text_size = font_metric.size(QtCore.Qt.TextExpandTabs, info)
        label.setFixedSize(text_size.width() + 50, text_size.height() + 30)
        v_layout.addWidget(label)
        h_layout = QtWidgets.QHBoxLayout()
        v_layout.addLayout(h_layout)
        name = QtWidgets.QLineEdit()
        name.setPlaceholderText(str(place_holder_text))
        name.setText(str(place_holder_text))
        create_button = QtWidgets.QPushButton('Create!')
        h_layout.addWidget(name)
        h_layout.addWidget(create_button)

        def do_create():
            try:
                nxt.create_context(name.text(),
                                   interpreter_exe=interpreter_exe,
                                   context_graph=context_graph,
                                   exe_script_args=exe_script_args)
                pop_up.close()
            except (IOError, NameError) as e:
                info = str(e)
                msg = 'Failed to create context!'
                logger.error(info)
                nxt_editor.dialogs.NxtWarningDialog.show_message(msg,
                                                                 info=info)

        create_button.pressed.connect(do_create)
        pop_up.exec_()

    def get_global_actions(self):
        """Get a list of NxtActions with the WindowShortcut context
        :return: List of NxtActions
        """
        global_actions = []
        for action in self.get_all_nxt_actions():
            if action.shortcutContext() == QtCore.Qt.WindowShortcut:
                global_actions += [action]
        return global_actions

    def get_all_nxt_actions(self):
        """Get a list of all NxtActions via the NxtActionContainer objects
        :return: List of NxtActions
        """
        all_actions = []
        all_containers = self.findChildren(actions.NxtActionContainer)
        for container in all_containers:
            all_actions += container.actions()
        return all_actions

    def get_hotkey_map(self):
        """Get a map of NxtAction containers and their actions in an
        ordered dict where each key is a row row for a QAbstractTableModel.
        :return: OrderedDict
        """
        hotkeys = OrderedDict()
        # Action container objects in the order we wish to display them
        action_containers = [self.app_actions, self.alignment_actions,
                             self.display_actions, self.view_actions,
                             self.layer_actions, self.node_actions,
                             self.property_manager_actions,
                             self.node_comment_actions,
                             self.execute_actions, self.code_editor_actions]
        for container in action_containers:
            hotkeys[container.objectName()] = container.get_action_data()
        return hotkeys

    @property
    def view(self):
        return self.get_current_view()

    @property
    def model(self):
        if self.view:
            return self.view.model

    def new_tab(self, initial_stage=None, update=True):
        """Open a new graph view, optionally on a specific initial graph.
        Create necessary model, pass to nxt to connect graph to model,
        create tab for new file.
        :param initial_stage: Graph object to make new view of.
        :type initial_stage: nxt.core.Graph.Graph
        :param update: If true the different views will update
        :type update: bool
        """
        self.set_waiting_cursor(True)
        # get new graph
        if initial_stage:
            stage = initial_stage
        else:
            stage = self.nxt.new_file()
        # create model
        model = StageModel(stage=stage)
        model.processing.connect(self.set_waiting_cursor)
        model.request_ding.connect(self.ding)
        model.layer_alias_changed.connect(partial(self.update_tab_title, model))
        # create view
        view = StageView(model=model, parent=self)
        # setup tab
        tab_index = self.open_files_tab_widget.count()
        self.open_files[model.uid] = {'stage': stage, 'model': model,
                                      'view': view}
        self.open_files_tab_widget.addTab(view, stage._name)
        if update:
            self.open_files_tab_widget.setCurrentIndex(tab_index)
            self.layer_manager.set_stage_model(model)
            model.layer_color_changed.connect(self.update_target_color)
            model.target_layer_changed.connect(self.update_target_color)
            model.comp_layer_changed.connect(self.update_target_color)
            self.update_target_color()
            self.update()  # TODO: Make this better
        self.set_waiting_cursor(False)

    @staticmethod
    def ding():
        if user_dir.user_prefs.get(user_dir.USER_PREF.DING, True):
            QtWidgets.QApplication.instance().beep()

    def center_view(self):
        target_graph_view = self.get_current_view()
        if target_graph_view:
            target_graph_view.centerOn(0, 0)

    def load_file(self, filepath=None):
        """Open an NxtFileDialog to allow user to select .nxt file to open.
        Attempt to open resulting choice. If an attempt is made to open a
        file that is already open, we will just focus that  tab.
        :param filepath: path to file on disk
        :type filepath: str

        :return: bool -- whether or not the file was successfully loaded.
        :rtype: bool
        """
        if not filepath:
            # TODO: The dialog should register the last opened folder into the
            #  user_dir and use that as the starting dir for file dialogs
            #  that aren't intrinicly tied to a layers real_path
            real_path = None
            try:
                real_path = self.model.stage.top_layer.real_path
            except AttributeError:
                pass
            _dir = os.path.dirname(real_path or os.getcwd())
            potential_path = NxtFileDialog.system_file_dialog(base_dir=_dir)
            if not potential_path:
                logger.debug("No file selected to load.")
                return
        else:
            potential_path = filepath
        # Try to load the file path via nxt
        self.set_waiting_cursor(True)
        new_stage = None
        try:
            new_stage = self.nxt.load_file(potential_path)
        except IOError as e:
            NxtWarningDialog.show_message("Failed to Open", str(e))
        self.set_waiting_cursor(False)
        if new_stage:
            self.new_tab(initial_stage=new_stage, update=not self.in_startup)
            user_dir.editor_cache[user_dir.USER_PREF.LAST_OPEN] = potential_path

    def save_open_tab(self):
        """Save the file that corresponds to the currently selected tab."""
        self.nxt.save_file(self.get_current_tab_file_path())

    def save_all_layers(self):
        if not self.model:
            return
        for layer in self.model.stage._sub_layers:
            self.save_layer(layer)

    def save_layer(self, layer=None):
        if not layer:
            layer = self.model.target_layer
        if not layer:
            return
        if not layer.real_path:
            layer_saved = self.save_layer_as(layer, open_in_new_tab=False)
        else:
            self.set_waiting_cursor(True)
            self.nxt.save_layer(layer)
            layer_saved = True
            user_dir.editor_cache[user_dir.USER_PREF.LAST_OPEN] = layer.real_path
        self.view.update_filepath()
        if layer_saved:
            try:
                self.model.effected_layers.remove(layer.real_path)
            except KeyError:  # Layer may not have been changed
                pass
        self.model.layer_saved.emit(layer.real_path)
        self.set_waiting_cursor(False)

    def save_layer_as(self, layer=None, open_in_new_tab=True):
        """Prompt the user to select a save location for a layer.

        If no layer is specified the target layer is used.

        :param layer: path to prompt for, defaults to target layer.
        :type layer: nxt_layer.SpecLayer, optional
        :param open_in_new_tab: If True, open the layer in a new tab after save
        :type start: str, optional
        :return: True if save was successful, false otherwise.
        :rtype: bool
        """
        if not layer:
            layer = self.model.target_layer
        old_real_path = layer.real_path
        old_path = layer.filepath
        if not old_real_path:
            open_in_new_tab = False
            base_dir = os.path.join(user_dir.USER_DIR, layer.get_alias())
        else:
            base_dir = layer.real_path
        caption = 'Save "{}"'.format(layer.get_alias())
        save_path = NxtFileDialog.system_file_dialog(base_dir, 'save',
                                                     caption=caption)
        if not save_path:
            return False
        self.set_waiting_cursor(True)
        self.nxt.save_layer(layer, filepath=save_path)
        user_dir.editor_cache[user_dir.USER_PREF.LAST_OPEN] = layer.real_path
        layer.filepath = old_path
        if open_in_new_tab:
            self.load_file(save_path)
            layer.real_path = old_real_path
        elif layer is self.model.top_layer:
            tab_idx = self.open_files_tab_widget.currentIndex()
            self.open_files_tab_widget.setTabText(tab_idx, layer.alias)
        tab_idx = self.open_files_tab_widget.currentIndex()
        self.on_tab_change(tab_idx)
        self.set_waiting_cursor(False)
        return True

    def open_source(self, layer):
        if not layer:
            layer = self.model.display_layer
        self.load_file(layer.real_path)

    def find_startpoint(self):
        """Cycles through start points"""
        if not self.model:
            return
        start_nodes = self.model.get_start_nodes()
        start_node_len = len(start_nodes)
        if not start_node_len:
            logger.warning("No start nodes found.")
            return
        in_range = self.last_focused_start in range(start_node_len)
        if in_range:
            idx = self.last_focused_start
        else:
            idx = 0
            self.last_focused_start = idx
        self.last_focused_start += 1
        self.model.select_and_frame(start_nodes[idx])

    def align_left(self):
        logger.info('align left')

    def align_hcenter(self):
        logger.info('align hcenter')

    def align_right(self):
        logger.info('align right')

    def align_top(self):
        logger.info('align top')

    def align_vcenter(self):
        logger.info('align vcenter')

    def align_bottom(self):
        logger.info('align bottom')

    def distribute_horizontal(self):
        logger.info('distribute horizontal')

    def distribute_vertical(self):
        logger.info('distribute vertical')

    def undo(self):
        current_view = self.get_current_view()
        if current_view:
            model = current_view.model
            model.undo()

    def redo(self):
        current_view = self.get_current_view()
        if current_view:
            model = current_view.model
            model.redo()

    def refresh_log_button(self):
        if self.output_log.isVisible():
            self.log_button.setText("Hide Log")
        else:
            self.log_button.setText("Show Log")

    def log_button_clicked(self):
        if self.output_log.isVisible():
            self.output_log.hide()
            return
        self.output_log.show()
        self.output_log.raise_()

    def update_tab_title(self, model, layer_changed):
        tab_idx = self.open_files_tab_widget.currentIndex()
        view = self.open_files_tab_widget.widget(tab_idx)
        cur_model = view.model
        if model is not cur_model:
            return
        if layer_changed != model.top_layer.real_path:
            return
        new_title = model.get_layer_alias(layer_changed)
        self.open_files_tab_widget.setTabText(tab_idx, new_title)

    def on_tab_change(self, tab_index):
        """Happens every tab change. Used to keep the dock widgets aware of
        the current graph model.
        """
        view = self.open_files_tab_widget.widget(tab_index)
        if not view:
            return
        if view == self.previous_view:
            return
        self.previous_view = view
        uid = view.model.uid
        self.last_focused_start = 0
        if uid in self.open_files.keys():
            model = self.open_files[uid]['model']
            layer_path = model.get_layer_path(model.top_layer)
            title = model.get_layer_alias(layer_path)
            self.open_files_tab_widget.setTabText(tab_index, title)
            self.property_editor.set_stage_model(model)
            self.code_editor.set_stage_model(model)
            self.layer_manager.set_stage_model(model)
            self.history_view.set_stage_model(model)
            self.workflow_tools.set_stage_model(model)
            self.build_view.set_stage_model(model)
            self.find_rep.set_stage_model(model)
            self.output_log.set_stage_model(model)
            self.update_target_color()
            logger.debug("Successfully set up new tab.")
            self.last_focused_tab = tab_index
            self.update_implicit_action()
            view.toggle_grid(
                user_dir.user_prefs.get(user_dir.USER_PREF.SHOW_GRID, True)
            )
            model.destroy_cmd_port.connect(self.update_cmd_port_action)
        else:
            logger.critical("Failed to set up new tab.")
        view.setFocus()
        self.tab_changed.emit()

    def get_current_tab_file_path(self):
        """Get the file path of the currently open tab.

        :return: File path of the currently open tab.
        :rtype:str
        """
        if not self.model:
            return
        return self.model.get_layer_path(self.model.stage.top_layer)

    def get_current_tab_model(self):
        """Get the file path of the currently open tab.
        :return: File path of the currently open tab.
        """
        idx = self.open_files_tab_widget.currentIndex()
        widget = self.open_files_tab_widget.widget(idx)
        if widget:
            uid = widget.model.uid
            return self.open_files[uid]['model']

    def get_current_view(self):
        return self.open_files_tab_widget.currentWidget()

    def update_cmd_port_action(self):
        self.execute_actions.enable_cmd_port_action.blockSignals(True)
        if self.model:
            state = self.model.use_cmd_port
        else:
            state = False
        self.execute_actions.enable_cmd_port_action.setChecked(state)
        self.execute_actions.enable_cmd_port_action.blockSignals(False)

    def update_implicit_action(self):
        self.view_actions.implicit_action.blockSignals(True)
        state = self.model.implicit_connections
        self.view_actions.implicit_action.setChecked(state)
        self.view_actions.implicit_action.blockSignals(False)

    def update_target_color(self):
        disp_layer = self.model.display_layer
        color = self.model.get_layer_color(disp_layer)

        # update widgets
        self.open_files_tab_widget.setStyleSheet('padding: 1; border: 1px solid %s' % color)
        self.open_files_tab_widget.update()
        self.code_editor.update_border_color()
        self.property_editor.update_styles()

    def keyPressEvent(self, event):
        key = event.key()
        if key not in self._held_keys:
            self._held_keys.append(key)
        self.zoom_keys_down = False
        match = QtGui.QKeySequence(*self._held_keys).matches(self.zoom_keys)
        if match == QtGui.QKeySequence.SequenceMatch.ExactMatch:
            self.zoom_keys_down = True
        event.accept()

    def keyReleaseEvent(self, event):
        key = event.key()
        if key in self._held_keys:
            self._held_keys.remove(key)
        self.zoom_keys_down = False
        match = QtGui.QKeySequence(*self._held_keys).matches(self.zoom_keys)
        if match == QtGui.QKeySequence.SequenceMatch.ExactMatch:
            self.zoom_keys_down = True

    def eventFilter(self, widget, event):
        # enter editing after update_code_is_local
        if event.type() == QtCore.QEvent.MouseButtonDblClick:
            if isinstance(widget.parent(), NxtCodeEditor):
                self.code_editor.update_code_is_local()
                self.code_editor.enter_editing()

        return False

    def show(self):
        """Centering after the window is shown because the center is based on the window's size."""
        # Todo: add previous rect to the bookmarks data for the layer - use this instead of center if it exists
        super(MainWindow, self).show()
        self.center_view()

    def showEvent(self, event):
        if self.state_last_hidden:
            self.restoreState(self.state_last_hidden)
            super(MainWindow, self).showEvent(event)
            return
        state_key = user_dir.EDITOR_CACHE.WINODW_STATE
        geo_key = user_dir.EDITOR_CACHE.MAIN_WIN_GEO
        saved_state = user_dir.editor_cache.get(state_key)
        if saved_state:
            self.restoreState(QtCore.QByteArray(saved_state))
        saved_geo = user_dir.editor_cache.get(geo_key)
        if saved_geo:
            self.restoreGeometry(QtCore.QByteArray(saved_geo))
        state_key = user_dir.EDITOR_CACHE.NODE_PROPERTY_STATE
        property_state = user_dir.editor_cache.get(state_key)
        if property_state:
            self.property_editor.model.state = property_state
        if self.view:
            self.view.setFocus()
        super(MainWindow, self).showEvent(event)

    def hideEvent(self, event):
        self.state_last_hidden = self.saveState()
        super(MainWindow, self).hideEvent(event)

    def closeEvent(self, event):
        """Check for unsaved work before accepting the event. If the event
        is accepted we also save the state of the UI before closing."""
        if self._closing:
            self._closing = False
            event.ignore()
            return
        dirty_models = []
        for open_file_dict in self.open_files.values():
            unsaved = open_file_dict['model'].get_unsaved_changes()
            if unsaved:
                dirty_models += [open_file_dict['model']]
        if dirty_models:
            resp = UnsavedLayersDialogue.save_before_exit(dirty_models, self)
            if resp == QtWidgets.QDialog.Rejected:
                event.ignore()
                return
        event.accept()
        self.shutdown_rpc_server()
        # Window state
        state_key = user_dir.EDITOR_CACHE.WINODW_STATE
        geo_key = user_dir.EDITOR_CACHE.MAIN_WIN_GEO
        user_dir.editor_cache[state_key] = self.saveState()
        user_dir.editor_cache[geo_key] = self.saveGeometry()
        state_key = user_dir.EDITOR_CACHE.NODE_PROPERTY_STATE
        property_state = self.property_editor.model.state
        if property_state:
            user_dir.editor_cache[state_key] = str(property_state)

        nxt_log.stop_session_log(self.nxt.log_file)
        # Close our dock widgets.
        for child in self.children():
            if isinstance(child, DockWidgetBase):
                child.close()
        # Save closing session
        closing_session = []
        for file_dict in self.open_files.values():
            model = file_dict['model']
            real_path = model.top_layer.real_path
            if not real_path:
                continue
            closing_session += [str(real_path)]
        if closing_session:
            pref_key = user_dir.EDITOR_CACHE.LAST_CLOSED
            last_sessions = user_dir.editor_cache.get(pref_key, [])
            last_sessions += [closing_session]
            user_dir.editor_cache[pref_key] = last_sessions
        self._closing = True
        self.close_signal.emit()
        super(MainWindow, self).closeEvent(event)

    def validate_layers_saved(self, model=None, single_layer=None):
        model = model or self.model
        if single_layer:
            layers = [single_layer]
        else:
            layers = model.stage._sub_layers
        unsaved = model.get_unsaved_changes(layers=layers)
        if unsaved and not single_layer:
            resp = UnsavedLayersDialogue.save_before_exit([model], self)
            if resp == QtWidgets.QDialog.Rejected:
                return False
        elif unsaved and single_layer:
            info = 'Layer "{}" has unsaved changes!'.format(single_layer.alias)
            resp = UnsavedChangesMessage.save_before_close(info=info)
            save = UnsavedChangesMessage.Save
            cancel = UnsavedChangesMessage.Cancel
            if resp == cancel:
                return False
            if resp == save:
                self.save_layer(single_layer)
                return True
        return True


class ToolBar(QtWidgets.QToolBar):

    def __init__(self, parent=None, resize_signal=None):
        super(ToolBar, self).__init__(parent=parent)
        # self.setFixedHeight(32)
        self.default_size = 19
        self.size = self.default_size
        self.setIconSize(QtCore.QSize(self.default_size, self.default_size))
        if resize_signal:
            resize_signal.connect(self.font_size_changed)

    def font_size_changed(self, delta):

        if delta == 0:
            self.size = self.default_size
        else:
            self.size = self.size + delta
        for action in self.actions():
            icon = action.icon()
            if icon.isNull():
                continue
            pixmap = icon.pixmap(self.size)
            new_pixmap = pixmap.scaled(QtCore.QSize(self.size, self.size),
                                       QtCore.Qt.IgnoreAspectRatio,
                                       QtCore.Qt.SmoothTransformation)

            action.setIcon(QtGui.QIcon(new_pixmap))

        self.setIconSize(QtCore.QSize(self.size, self.size))
        self.update()


class NodeAuthoringToolBar(ToolBar):

    def __init__(self, parent=None, resize_signal=None):
        super(NodeAuthoringToolBar, self).__init__(parent=parent,
                                                   resize_signal=resize_signal)
        self.setObjectName('Node Authoring')
        self.main_window = parent
        self.node_actions = self.main_window.node_actions
        self.main = QtWidgets.QWidget()
        self.addWidget(self.main)

        self.layout = QtWidgets.QGridLayout()
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setSpacing(0)
        self.main.setLayout(self.layout)
        # add node
        self.addAction(self.node_actions.add_node_action)
        # delete node
        self.addAction(self.node_actions.delete_node_action)
        self.addSeparator()
        # duplicate node
        self.addAction(self.node_actions.duplicate_node_action)
        # instance node
        self.addAction(self.node_actions.instance_node_action)
        # remove instance node
        self.addAction(self.node_actions.remove_instance_action)
        self.addSeparator()
        # cut node
        self.addAction(self.node_actions.cut_node_action)
        # copy node
        self.addAction(self.node_actions.copy_node_action)
        # paste node
        self.addAction(self.node_actions.paste_node_action)
        self.addSeparator()

        # localize node
        self.addAction(self.node_actions.localize_node_action)
        # revert node
        self.addAction(self.node_actions.revert_node_action)
        self.addSeparator()
        # select all
        self.addAction(self.node_actions.select_all_action)


class AlignDistributeToolBar(ToolBar):

    def __init__(self, parent=None, resize_signal=None):
        super(AlignDistributeToolBar, self).__init__(parent=parent,
                                                     resize_signal=resize_signal)
        self.setObjectName('Alignment Tools')
        self.main_window = parent
        # ACTIONS
        self.addActions(self.main_window.alignment_actions.actions())


class ExecuteToolBar(ToolBar):

    def __init__(self, parent=None, resize_signal=None):
        super(ExecuteToolBar, self).__init__(parent=parent,
                                             resize_signal=resize_signal)
        self.setObjectName('Execute Tools')
        self.main_window = parent
        self.exec_actions = self.main_window.execute_actions
        self.addActions([self.exec_actions.execute_graph_action,
                         self.exec_actions.stop_exec_action,
                         self.exec_actions.execute_selected_action,
                         self.exec_actions.execute_from_action,
                         self.exec_actions.execute_hierarchy_action])
        self.addSeparator()
        self.addActions([self.exec_actions.add_start_action,
                         self.exec_actions.remove_start_action,
                         self.exec_actions.find_start_action])
        self.addSeparator()
        self.addActions([self.exec_actions.add_break_action,
                         self.exec_actions.remove_break_action,
                         self.exec_actions.clear_breaks_action])


class DisplayToolBar(ToolBar):

    def __init__(self, parent=None, resize_signal=None):
        super(DisplayToolBar, self).__init__(parent=parent,
                                             resize_signal=resize_signal)
        self.setObjectName('Display Tools')
        self.main_window = parent
        self.view_actions = self.main_window.view_actions
        self.display_actions = self.main_window.display_actions
        self.addAction(self.display_actions.raw_action)
        self.addAction(self.display_actions.resolve_action)
        self.addAction(self.display_actions.cached_action)
        self.addSeparator()
        # Connection view
        self.addAction(self.view_actions.grid_action)
        self.addAction(self.view_actions.implicit_action)
        self.addSeparator()
        self.addAction(self.view_actions.frame_all_action)
        self.addAction(self.view_actions.frame_selection_action)
        self.addSeparator()
        self.addAction(self.view_actions.hide_attrs_action)
        self.addAction(self.view_actions.disp_local_attrs_action)
        self.addAction(self.view_actions.disp_inst_attrs_action)
        self.addAction(self.view_actions.disp_all_attrs_action)


class MenuBar(QtWidgets.QMenuBar):

    """Menu bar for nxt main window"""

    def __init__(self, parent=None):
        super(MenuBar, self).__init__(parent=parent)
        self.main_window = parent
        self.app_actions = parent.app_actions  # type: actions.AppActions
        self.exec_actions = parent.execute_actions  # type: actions.ExecuteActions
        self.node_actions = parent.node_actions  # type: actions.NodeActions
        self.ce_actions = parent.code_editor_actions  # type: actions.CodeEditorActions
        self.display_actions = parent.display_actions  # type: actions.DisplayActions
        self.view_actions = parent.view_actions  # type: actions.StageViewActions
        self.layer_actions = parent.layer_actions  # type: actions.LayerActions
        # File Menu
        self.file_menu = self.addMenu('File')
        self.file_menu.setTearOffEnabled(True)
        # ACTIONS
        # Something of note:
        # Menu actions with multi key shortcuts are (in general) act like
        # application level shortcuts on OSX. Single key shortcuts however
        # do not work in the same way. There is a workaround for this if Qt
        # never fixes how menus are made and we get complaints from osx users.
        # https://thebreakfastpost.com/2014/06/03/single-key-menu-shortcuts-with-qt5-on-osx/
        # New tab
        self.file_menu.addAction(self.main_window.app_actions.new_graph_action)
        # Open file
        self.file_menu.addAction(self.main_window.app_actions.open_file_action)

        # Recent files
        self.load_recent_menu = RecentFilesMenu(action_target=self.main_window.load_file)
        self.file_menu.addMenu(self.load_recent_menu)
        self.file_menu.addAction(self.layer_actions.save_layer_action)
        self.file_menu.addAction(self.layer_actions.save_layer_as_action)
        self.file_menu.addSeparator()
        self.file_menu.addAction(self.layer_actions.save_all_layers_action)
        self.file_menu.addSeparator()
        self.file_menu.addAction(self.layer_actions.new_layer_above_action)
        self.file_menu.addAction(self.layer_actions.new_layer_below_action)
        self.file_menu.addSeparator()
        self.file_menu.addAction(self.layer_actions.ref_layer_above_action)
        self.file_menu.addAction(self.layer_actions.ref_layer_below_action)
        self.file_menu.addSeparator()
        self.builtins_menu = QtWidgets.QMenu('Reference Builtin Graph')
        self.builtins_menu.aboutToShow.connect(partial(populate_builtins_menu,
                                                       qmenu=self.builtins_menu,
                                                       main_window=self.main_window))
        self.file_menu.addMenu(self.builtins_menu)
        # Close app
        self.file_menu.addSeparator()
        self.file_menu.addAction(self.main_window.app_actions.close_tab_action)
        self.file_menu.addAction(self.main_window.app_actions.close_action)

        # Edit Menu
        self.edit_menu = self.addMenu('Edit')
        self.edit_menu.setTearOffEnabled(True)

        self.edit_menu.addAction(self.main_window.app_actions.undo_action)
        self.edit_menu.addAction(self.main_window.app_actions.redo_action)
        self.edit_menu.addSeparator()
        self.edit_menu.addAction(self.node_actions.copy_node_action)
        self.edit_menu.addAction(self.node_actions.cut_node_action)
        self.edit_menu.addAction(self.node_actions.paste_node_action)
        self.edit_menu.addAction(self.node_actions.delete_node_action)
        self.edit_menu.addSeparator()
        self.edit_menu.addAction(self.node_actions.select_all_action)

        # view menu
        self.view_menu = self.addMenu('View')
        self.view_menu.setTearOffEnabled(True)
        self.view_menu.addAction(self.view_actions.frame_selection_action)
        self.view_menu.addAction(self.view_actions.frame_all_action)
        self.view_menu.addSeparator()
        self.view_menu.addAction(self.display_actions.raw_action)
        self.view_menu.addAction(self.display_actions.resolve_action)
        self.view_menu.addAction(self.display_actions.cached_action)
        self.view_menu.addSeparator()
        self.view_menu.addAction(self.view_actions.implicit_action)
        self.view_menu.addAction(self.view_actions.grid_action)
        self.view_opt_menu = self.view_menu.addMenu('Options')
        self.view_opt_menu.setTearOffEnabled(True)
        self.view_opt_menu.addAction(self.view_actions.tooltip_action)
        self.view_opt_menu.addAction(self.layer_actions.lay_manger_table_action)
        self.view_opt_menu.addAction(self.ce_actions.overlay_message_action)
        self.view_opt_menu.addAction(self.app_actions.increase_font_size_action)
        self.view_opt_menu.addAction(self.app_actions.decrease_font_size_action)
        self.view_opt_menu.addAction(self.app_actions.reset_font_size_action)
        self.app_actions.increase_font_size_action.triggered.connect(
            self.main_window.increase_font_size)
        self.app_actions.decrease_font_size_action.triggered.connect(
            self.main_window.decrease_font_size)
        self.app_actions.reset_font_size_action.triggered.connect(
            self.main_window.reset_font_size)

        # graph menu
        self.graph_menu = self.addMenu('Graph')
        self.graph_menu.setTearOffEnabled(True)
        self.graph_menu.addAction(self.node_actions.add_node_action)

        # execute menu
        self.execute_menu = self.addMenu('Execute')
        self.execute_menu.setTearOffEnabled(True)
        self.execute_menu.addAction(self.exec_actions.execute_from_action)
        self.execute_menu.addAction(self.exec_actions.execute_selected_action)
        self.execute_menu.addAction(self.exec_actions.execute_hierarchy_action)
        self.execute_menu.addAction(self.exec_actions.execute_graph_action)
        self.execute_menu.addAction(self.exec_actions.clear_cache_action)
        self.execute_menu.addAction(self.exec_actions.wt_recomp_action)
        # Populate action data for window actions
        self.app_actions.layer_manager_action.setData(parent.layer_manager)
        self.app_actions.property_editor_action.setData(parent.property_editor)
        self.app_actions.code_editor_action.setData(parent.code_editor)
        self.app_actions.history_view_action.setData(parent.history_view)
        self.app_actions.build_view_action.setData(parent.build_view)
        self.app_actions.output_log_action.setData(parent.output_log)
        self.app_actions.hotkey_editor_action.setData(parent.hotkey_editor)
        self.app_actions.workflow_tools_action.setData(parent.workflow_tools)

        # window menu
        self.window_menu = self.addMenu('Window')
        self.window_menu.aboutToShow.connect(self.populate_window_menu)
        self.window_menu.triggered.connect(self.window_action_triggered)
        self.window_menu_actions = [
            self.app_actions.layer_manager_action,
            self.app_actions.property_editor_action,
            self.app_actions.code_editor_action,
            self.app_actions.history_view_action,
            self.app_actions.build_view_action,
            self.app_actions.output_log_action,
            self.app_actions.hotkey_editor_action,
            self.app_actions.workflow_tools_action
        ]
        self.populate_window_menu()
        # Remote Menu
        self.remote_menu = self.addMenu('Remote')
        remote_context_action = self.remote_menu.addAction('Create Remote '
                                                           'Context')
        remote_context_func = self.main_window.create_remote_context
        remote_context_action.triggered.connect(remote_context_func)
        if not is_standalone():
            remote_context_action.setEnabled(False)
        self.remote_menu.addSeparator()
        self.remote_menu.addAction(self.exec_actions.enable_cmd_port_action)
        self.remote_menu.addSeparator()
        self.remote_menu.addAction(self.exec_actions.startup_rpc_action)
        self.remote_menu.addAction(self.exec_actions.shutdown_rpc_action)
        self.options_menu = self.addMenu('Options')
        self.options_menu.addAction(self.app_actions.toggle_ding_action)
        self.options_view_sub = self.options_menu.addMenu('View')
        self.options_view_sub.setTearOffEnabled(True)
        self.options_view_sub.addActions(self.view_opt_menu.actions())
        # Help Menu
        self.help_menu = self.addMenu('Help')
        self.help_menu.setTearOffEnabled(True)
        prefs_dir_action = self.help_menu.addAction('Open Prefs Dir')
        prefs_dir_action.triggered.connect(self.open_prefs_dir)
        config_dir_action = self.help_menu.addAction('Open Plugins Dir')
        config_dir_action.triggered.connect(self.open_plugins_dir)
        self.help_menu.addSeparator()
        self.help_menu.addAction(self.main_window.app_actions.docs_action)
        github_action = self.help_menu.addAction('GitHub')
        url = 'https://github.com/nxt-dev/nxt_editor'
        github_action.triggered.connect(partial(webbrowser.open_new, url))
        self.help_menu.addSeparator()
        del_resources = self.help_menu.addAction('Clear UI Icon Cache')
        del_resources.triggered.connect(self.delete_resources_pyc)
        self.help_menu.addSeparator()
        # Secret Menu
        self.secret_menu = self.help_menu.addMenu('Developer Options')
        self.secret_menu.setTearOffEnabled(True)
        test_log_action = self.secret_menu.addAction('test logging')
        test_log_action.triggered.connect(self.__test_all_logging)
        print_action = self.secret_menu.addAction('test print')
        print_action.triggered.connect(self.__test_print)
        critical_action = self.secret_menu.addAction('test remove layer')
        critical_action.triggered.connect(self.__test_rm_layer)
        uncaught_exception = self.secret_menu.addAction('uncaught exception')
        uncaught_exception.triggered.connect(self.__force_uncaught_exception)
        compile_selection = self.secret_menu.addAction('compile selection')
        compile_selection.triggered.connect(self.__compile_node_code)
        save_cache = self.secret_menu.addAction('save cached')
        save_cache.triggered.connect(self.__save_cache_layer)
        load_cache = self.secret_menu.addAction('load cached')
        load_cache.triggered.connect(self.__load_cache_layer)
        rpc_ping = self.secret_menu.addAction('rpc ping')
        rpc_ping.triggered.connect(self.__rpc_ping)
        force_kill_rpc = self.secret_menu.addAction('force kill rpc')
        force_kill_rpc.triggered.connect(self.__force_kill_rpc)
        # Debugger function
        test_graph_action = self.secret_menu.addAction('Debugger')
        test_graph_action.triggered.connect(self.__debug)
        # Force redraw
        force_redraw_action = self.secret_menu.addAction(
            'Force Redraw')
        force_redraw_action.triggered.connect(self.__force_redraw)
        # Force rebuild stage
        force_build_stage_action = self.secret_menu.addAction('Force Update')
        force_build_stage_action.triggered.connect(self.__force_build_stage)
        self.help_menu.addSeparator()
        about_action = self.help_menu.addAction('About')
        about_action.triggered.connect(self.about_message)

    def eventFilter(self, widget, event):
        if event.type() == QtCore.QEvent.Type.ShortcutOverride:
            return True
        return False

    def populate_window_menu(self):
        self.window_menu.clear()
        for action in self.window_menu_actions:
            widget = action.data()
            action.setChecked(widget.isVisible())
            self.window_menu.addAction(action)
        self.window_menu.addSeparator()
        for file_dict in self.main_window.open_files.values():
            widget = file_dict['view']
            name = file_dict['model'].top_layer.get_alias()
            new_action = self.window_menu.addAction(name)
            new_action.setData(widget)

    def window_action_triggered(self, action=None):
        if not action:
            # Sometimes Qt sends us this signal with no action.
            return
        widget = action.data()
        tab_index = self.main_window.open_files_tab_widget.indexOf(widget)
        if tab_index != -1:
            self.main_window.open_files_tab_widget.setCurrentIndex(tab_index)
            return
        if action.isChecked():
            widget.show()
            widget.raise_()
        else:
            widget.close()

    @staticmethod
    def open_prefs_dir():
        d = user_dir.PREF_DIR
        if 'darwin' in sys.platform:
            os.system('open {}'.format(d))
        elif 'win' in sys.platform:
            os.startfile(d)
        else:
            try:
                os.system('xdg-open {}'.format(d))
            except:
                logger.exception('Failed to open user dir')

    @staticmethod
    def open_plugins_dir():
        d = USER_PLUGIN_DIR
        if 'darwin' in sys.platform:
            os.system('open {}'.format(d))
        elif 'win' in sys.platform:
            os.startfile(d)
        else:
            try:
                os.system('xdg-open {}'.format(d))
            except:
                logger.exception('Failed to open user config dir')

    def about_message(self):
        text = ('nxt {} \n'
                'graph v{}\n'
                'api v{}\n'
                'editor v{}\n'
                'Copyright (c) 2015-2020 '
                'The nxt Authors').format(self.main_window.host_app,
                                          GRAPH_VERSION.VERSION_STR,
                                          API_VERSION.VERSION_STR,
                                          EDITOR_VERSION.VERSION_STR)
        message_box = QtWidgets.QMessageBox()
        message_box.setWindowTitle('About nxt '
                                   '({})'.format(EDITOR_VERSION.VERSION_STR))
        message_box.setText(text)
        message_box.setStandardButtons(message_box.Close)
        message_box.setIcon(message_box.Icon.Information)
        message_box.exec_()

    @staticmethod
    def delete_resources_pyc():
        ui_dir = os.path.dirname(__file__)
        resources_file = os.path.join(ui_dir, 'qresources.py').replace(os.sep,
                                                                      '/')
        resources_file_c = os.path.join(ui_dir, 'qresources.pyc').replace(os.sep,
                                                                          '/')
        success = False
        if os.path.isfile(resources_file):
            try:
                os.remove(resources_file)
                success = True
            except:
                logger.exception('Failed to delete "{}" please do so '
                                 'manually.'.format(resources_file))
        if os.path.isfile(resources_file_c):
            try:
                os.remove(resources_file_c)
                success = True
            except:
                logger.exception('Failed to delete "{}" please do so '
                                 'manually.'.format(resources_file_c))
                success = False

        if success:
            logger.info('Cleared UI icon cache, please restart nxt.')
        from . import make_resources
        make_resources()

    def __test_print(self):
        """prints a simple message for output log debug"""
        print('Test print please ignore')

    def __test_all_logging(self):
        done = []
        for level_num in logging._levelNames:
            if not isinstance(level_num, int):
                level_num = logging.getLevelName(level_num)
            if level_num in done:
                continue
            done += [level_num]
            logger.log(level_num, 'Testing logger level '
                                  '{}'.format(logging.getLevelName(level_num)))

    def __test_rm_layer(self):
        nxt_object = self.parent().nxt
        stage_key = nxt_object._loaded_files.keys()[0]
        stage = nxt_object._loaded_files[stage_key]
        stage.remove_sublayer(1)
        model = self.parent().model
        model.update_comp_layer()

    def __force_build_stage(self):
        self.main_window.model.update_comp_layer(rebuild=True)

    def __force_redraw(self):
        view = self.parent().view
        view.update_view()

    def __force_uncaught_exception(self):
        print(foo)

    def __compile_node_code(self):
        """Test the compile of a node's compute and if it works in the console
        :return:
        """
        path = self.main_window.model.selection[0]
        comp_layer = self.main_window.model.comp_layer
        rt_layer = self.main_window.model.stage.setup_runtime_layer(comp_layer)
        rt_node = rt_layer.lookup(path)
        from runtime import GraphError, Console
        import nxt.stage as _stage
        g = {'__stage__': self.main_window.model.stage,
             'STAGE': rt_layer,
             'w': _stage.w,
             }
        func = self.main_window.model.stage.get_node_code(rt_node,
                                                                  rt_layer)
        console = Console(g, node_path=path)
        g['func'] = func
        g['self'] = rt_node
        try:
            console.runcode(func)
        except GraphError:
            pass

    def __debug(self):
        model = self.main_window.model
        stage = model.stage
        target_layer = model.target_layer
        comp_layer = model.comp_layer
        nxt_object = self.parent().nxt
        stages = []
        layers = []
        for k in nxt_object._loaded_files.keys():
            stages.append(nxt_object._loaded_files[k])
        for stage in stages:
            stage.debug = True
            for l in stage._sub_layers:
                layers.append(l)
        return

    def __load_cache_layer(self):
        filt = 'nxt files (*.nxt)'
        file_path = QtWidgets.QFileDialog.getOpenFileName(filter=filt)[0]
        if not file_path:
            return
        layer_data = nxt_io.load_file_data(file_path)
        model = self.main_window.model
        cache_layer = nxt_layer.CacheLayer.load_from_layer_data(layer_data)
        if not model.current_rt_layer:
            model.current_rt_layer = nxt_layer.CompLayer()
        model.current_rt_layer.cache_layer = cache_layer

    def __save_cache_layer(self):
        curr_rt = self.main_window.model.current_rt_layer
        if not curr_rt:
            logger.info("No cache data to save")
            return
        filt = 'nxt files (*.nxt)'
        file_path = QtWidgets.QFileDialog.getSaveFileName(filter=filt)[0]
        if not file_path:
            return
        curr_rt.cache_layer.save(file_path)

    def __rpc_ping(self):
        proxy = NxtClient()
        proxy.is_alive()

    def __force_kill_rpc(self):
        proxy = NxtClient()
        proxy.kill()


class OpenFilesTabWidget(QtWidgets.QTabWidget):
    def __init__(self, parent=None):
        super(OpenFilesTabWidget, self).__init__(parent=parent)
        self.main_window = parent
        self.setTabsClosable(True)
        self.setMovable(True)
        self.tabCloseRequested.connect(self.close_tab)

    def close_tab(self, index):
        model = self.widget(index).model
        safe_to_close = self.main_window.validate_layers_saved(model=model)
        if not safe_to_close:
            self.main_window.set_waiting_cursor(False)
            return
        self.main_window.set_waiting_cursor(True)
        self.widget(index).clear()
        uid = self.widget(index).model.uid
        self.parent().nxt.unload_file(uid)
        tab_data = self.parent().open_files.pop(uid)
        model = tab_data['model']
        view = tab_data['view']
        view.deleteLater()
        model.deleteLater()
        self.removeTab(index)
        self.main_window.tab_changed.emit()
        real_path = model.top_layer.real_path
        if not real_path:
            self.main_window.set_waiting_cursor(False)
            return
        pref_key = user_dir.EDITOR_CACHE.LAST_CLOSED
        last_sessions = user_dir.editor_cache.get(pref_key, [])
        last_sessions += [[str(real_path)]]
        user_dir.editor_cache[pref_key] = last_sessions
        self.main_window.set_waiting_cursor(False)


class RecentFilesMenu(QtWidgets.QMenu):
    def __init__(self, action_target=None):
        super(RecentFilesMenu, self).__init__('Open Recent')
        self.aboutToShow.connect(self.refresh_list)
        self.action_target = action_target
        self.triggered.connect(self.recent_selected)

    def refresh_list(self):
        self.clear()
        recents = user_dir.editor_cache.get(user_dir.USER_PREF.RECENT_FILES, [])
        if not recents:
            action = self.addAction('No recents found')
            action.setEnabled(False)
        for file_path in recents:
            self.addAction(str(file_path))

    def recent_selected(self, action):
        self.action_target(action.text())


class StatusBarHandler(logging.Handler):
    def __init__(self, output_log=None):
        logging.Handler.__init__(self, level=logging.DEBUG)
        self.output_template = "{level} | {module}: \"{message}\""
        self.output_log = output_log
        self.signaller = LoggingSignaler()
        self.signaller.signal.connect(self.update)

    def emit(self, record):
        return
        self.signaller.signal.emit(record)

    def update(self, record):
        out_message = self.output_template.format(level=record.levelname,
                                                  module=record.module,
                                                  message=record.getMessage())
        if self.output_log:
            self.output_log.showMessage(out_message)


class StartRPCThread(QtCore.QThread):
    def __init__(self, main_window):
        super(StartRPCThread, self).__init__()
        self.main_window = main_window

    def run(self):
        if self.main_window.model:
            self.main_window.model.processing.emit(True)
        # We setup the log file here so we're tailing it _before_ we start
        # the server up.
        rpc_log = nxt_io.generate_temp_file(suffix='.nxtlog')
        # Setup rpc server log tail
        self.main_window.safe_stop_rpc_tailing()
        self.main_window.rpc_log_tail = FileTailingThread(rpc_log)
        self.main_window.handle_rpc_tailing_signals(True)
        self.main_window.rpc_log_tail.start()
        sh = QtLogStreamHandler.get_handler(self.main_window.new_log_signal)
        try:
            self.main_window.nxt._start_rpc_server(custom_stdout=True,
                                                   rpc_log_filepath=rpc_log,
                                                   socket_log=True,
                                                   stream_handler=sh)
        except OSError:
            logger.warning('Failed to start/connect to rpc server. Please try '
                           'starting the rpc server via the UI')
            if self.main_window.model:
                self.main_window.model.processing.emit(False)
            return
        remote_rpc_log_file_path = None
        if not self.main_window.nxt.rpc_server:
            proxy = NxtClient()
            try:
                remote_rpc_log_file_path = proxy.get_log_location()
            except:
                logger.warning('Failed to tail remote rpc server log!')
        if remote_rpc_log_file_path:
            self.main_window.rpc_log_tail.watch_path = remote_rpc_log_file_path
            with open(remote_rpc_log_file_path, 'r') as fp:
                text = fp.read()
                end_pos = len(text)
            self.main_window.rpc_log_tail.last_read_pos = end_pos
        if self.main_window.model:
            self.main_window.model.processing.emit(False)


def populate_builtins_menu(qmenu, main_window, layer=None):
    """Populates a QMenu object with actions for referencing each builtin layer.
    :param qmenu: QMenu object to be filled with actions
    :param main_window: nxt MainWindow
    :param layer: Optional layer to reference builtin layer under, if none is
    supplied the target layer is used.
    :return: QMenu
    """
    qmenu.clear()
    stage_model = main_window.model
    if not stage_model:
        enable = False
        idx = -1
    else:
        enable = True
        layer = layer or stage_model.target_layer
        idx = layer.layer_idx() + 1

    for file_name in os.listdir(nxt_io.BUILTIN_GRAPHS_DIR):
        if not file_name.endswith('.nxt'):
            continue
        new_action = qmenu.addAction(file_name)
        path = '${var}/{file_name}'.format(var=nxt_io.BUILTIN_GRAPHS_ENV_VAR,
                                           file_name=file_name)
        if enable:
            new_action.triggered.connect(partial(stage_model.reference_layer,
                                                 path, idx))
        new_action.setEnabled(enable)
    return qmenu


def nxt_execpthook(typ, value, tb):
    if 'nxt' not in tb.tb_frame.f_code.co_filename:
        return og_excepthook(typ, value, tb)
    logger.error('NXT encountered an Uncaught exception!')
    traceback.print_tb(tb)
    message = ('Please copy the error details and send to an nxt '
               'developer.\n'
               'Save your work immediately.')
    # TODO: Get the last few lines from the session log and put them here
    details = ''.join(traceback.format_exception(typ, value, tb))
    logger.exception(details)
    style_file = QtCore.QFile(':styles/styles/dark/dark.qss')
    style_file.open(QtCore.QFile.ReadOnly)
    stylesheet = str(style_file.readAll())
    dialog = NxtWarningDialog('Uncaught Exception!', message, details)
    dialog.setStyleSheet(stylesheet)
    dialog.exec_()


def catch_exceptions():
    debugger_attached = 'pydevd' in sys.modules
    return not debugger_attached


if sys.excepthook is not nxt_execpthook:
    og_excepthook = sys.excepthook

if catch_exceptions():
    sys.excepthook = nxt_execpthook



# Built-in
import re
import logging

# External
from Qt import QtCore
from Qt.QtWidgets import QDockWidget

# Internal
import nxt_editor

logger = logging.getLogger(nxt_editor.LOGGER_NAME)


class DockWidgetBase(QDockWidget):

    """All of nxt's dock widgets have a lot in common, this thin class
    mutates QDockWidget to nxt preferences"""

    def __init__(self, title, parent=None, graph_model=None,
                 minimum_width=None, minimum_height=None):
        # We have to prefix the object name with Nxt so we don't blow up our
        # host context.
        safe_name = 'Nxt ' + title
        super(DockWidgetBase, self).__init__(safe_name, parent=parent)
        self.setFeatures(QDockWidget.DockWidgetClosable |
                         QDockWidget.DockWidgetMovable |
                         QDockWidget.DockWidgetFloatable)
        self.setFocusPolicy(QtCore.Qt.FocusPolicy.ClickFocus)
        # set minimum dimensions
        if minimum_width:
            self.setMinimumWidth(minimum_width)

        if minimum_height:
            self.setMinimumHeight(minimum_height)

        # set title
        self.title = title
        self.setWindowTitle(title)
        # keep only a-zA-Z0-9 for object name
        self.setObjectName(re.sub(r'[^a-zA-Z0-9]', '', safe_name))

        # set title size
        self.setStyleSheet('''QDockWidget{
                                font-size: 7pt;
                                font-weight: bold;
                                color: grey;
                              }

                              QDockWidget::title{
                                text-align: center;
                                color: grey;
                              }
                              ''')

        # set graph model
        self.stage_model = graph_model
        self.model_signal_connections = []
        self.topLevelChanged.connect(self.on_window_status_changed)

    def set_stage_model(self, stage_model):
        """Sets the stage model for docwidgets to use, also calls the
        connection handling method.
        :param stage_model: StageModel
        """
        if self.stage_model:
            self.set_stage_model_connections(self.stage_model, False)
        self.stage_model = stage_model
        if self.stage_model:
            self.set_stage_model_connections(self.stage_model, True)

    def on_window_status_changed(self, is_window):
        if is_window:
            flags = (
                QtCore.Qt.Window |
                QtCore.Qt.CustomizeWindowHint |
                QtCore.Qt.WindowMinMaxButtonsHint |
                QtCore.Qt.WindowCloseButtonHint
            )
            self.setWindowFlags(flags)
            self.show()

    def set_stage_model_connections(self, model, connect):
        """Connect and disconnect signals from the docwidget.
        Called by `set_graph_model`.
        Sub-classes can add their own custom connection mapping before
        supering this method. Example:
            self.model_signal_connections = [
                (model.a_signal, self.method_to_be_called),
                (model.node_changed, self.update_title)
            ]
            super(SomeDoc, self).set_stage_model_connection(model, connect)
        :param model: Model to change connection status to.
        :type model: StageModel
        :param connect: True if connecting, False if disconnecting
        :type connect: bool
        """
        self.model_signal_connections += [
            (model.destroyed, self.on_stage_model_destroyed)
        ]
        for model_signal, my_func in self.model_signal_connections:
            if connect:
                model_signal.connect(my_func)
            else:
                model_signal.disconnect(my_func)
        self.model_signal_connections = []

    def on_stage_model_destroyed(self):
        """Called when the model's `destroyed` signal is emitted. Should be
        overloaded on sub-classes as its a good place to hide yo kids so users
        can't accidentally get to them after the model is destroyed.
        """
        self.stage_model = None



# Builtin
from functools import partial

# External
from Qt import QtWidgets, QtGui, QtCore

# Internal
from nxt.nxt_path import NODE_SEP


class FinderLineEdit(QtWidgets.QLineEdit):

    focus_node = QtCore.Signal(str)

    def __init__(self, parent, node_paths=()):
        super(FinderLineEdit, self).__init__(parent)
        completer = QtWidgets.QCompleter(node_paths, parent=self)
        completer.activated.connect(self.focus_node.emit)
        completer.setCaseSensitivity(QtCore.Qt.CaseInsensitive)
        completer.popup().setStyleSheet(parent.parent().stylesheet)
        self.returnPressed.connect(partial(self.focus_node.emit, None))
        self.setCompleter(completer)
        self.setFocusPolicy(QtCore.Qt.FocusPolicy.StrongFocus)

    def focusOutEvent(self, event):
        super(FinderLineEdit, self).focusOutEvent(event)
        self.parent().close()

    def keyPressEvent(self, event):
        super(FinderLineEdit, self).keyPressEvent(event)
        if event.key() == QtCore.Qt.Key_Escape:
            self.setText('')
            self.parent().close()


class FinderWidget(QtWidgets.QWidget):
    def __init__(self, parent, stage_model):
        super(FinderWidget, self).__init__(parent, QtCore.Qt.FramelessWindowHint)
        self.setAttribute(QtCore.Qt.WA_TranslucentBackground)
        self.stage_model = stage_model
        self.setFixedHeight(64)
        self.setFixedWidth(256)
        self.layout = QtWidgets.QVBoxLayout(self)
        node_paths = self.stage_model.comp_layer.descendants(include_implied=1)
        # Could use nxt_layer.sort_node_data but its too slow
        node_paths = sorted(node_paths)
        self.line_edit = FinderLineEdit(self, node_paths)
        self.line_edit.focus_node.connect(self.select_and_frame)
        self.layout.addWidget(self.line_edit)
        self.line_edit.setStyleSheet(self.parent().stylesheet)

    def show(self):
        super(FinderWidget, self).show()
        global_pos = QtGui.QCursor.pos()
        local_pos = self.mapFromGlobal(global_pos)
        self.move(local_pos)
        self.line_edit.setFocus()
        self.line_edit.setText(NODE_SEP)
        self.line_edit.completer().setCompletionPrefix(NODE_SEP)
        self.line_edit.completer().complete()

    def select_and_frame(self, node_path=None):
        if not node_path:
            node_path = self.line_edit.text()
        if node_path:
            self.stage_model.select_and_frame(node_path)
        self.close()



# Built-in
import math
import logging
import textwrap
import sys
from collections import OrderedDict

# External
from Qt import QtWidgets
from Qt import QtGui
from Qt import QtCore
from PySide2 import __version_info__ as qt_version

# Internal
import nxt_editor
from nxt import nxt_path, nxt_node
from nxt.nxt_layer import LAYERS
from . import colors
from nxt.stage import INTERNAL_ATTRS
from .label_edit import NameEditDialog
from .user_dir import USER_PREF, user_prefs


logger = logging.getLogger(nxt_editor.LOGGER_NAME)

MIN_LOD = user_prefs.get(USER_PREF.LOD, .4)

_pyside_version = qt_version


if _pyside_version[1] < 11:
    graphic_type = QtWidgets.QGraphicsItem
else:
    graphic_type = QtWidgets.QGraphicsObject


class NodeGraphicsItem(graphic_type):
    """The graphics item used to represent nodes in the graph. Contains
    instances of NodeGraphicsPlug for each attribute on the associated node.
    Contains functionality for arranging children into stacks.
    """

    ATTR_PLUG_RADIUS = 4
    EXEC_PLUG_RADIUS = 6
    ROUND_X = 2.
    ROUND_Y = 2.

    def __init__(self, model, node_path, view):
        super(NodeGraphicsItem, self).__init__()
        self.count = 0
        self.dim_factor = 100
        self.colors = [QtGui.QColor(QtCore.Qt.darkGray)]
        self.color_alpha = 1.0
        # graph settings
        self.node_path = node_path
        self.model = model
        self.view = view

        # item settings
        self.setCacheMode(QtWidgets.QGraphicsItem.DeviceCoordinateCache)

        self.setFlags(QtWidgets.QGraphicsItem.ItemIsMovable |
                      QtWidgets.QGraphicsItem.ItemIsFocusable |
                      QtWidgets.QGraphicsItem.ItemIsSelectable |
                      QtWidgets.QGraphicsItem.ItemSendsScenePositionChanges |
                      QtWidgets.QGraphicsItem.ItemNegativeZStacksBehindParent)
        self.setAcceptHoverEvents(True)

        # draw settings
        self.title_font = QtGui.QFont("Roboto Mono", 14)
        self.attr_font = QtGui.QFont("Roboto Mono", 9)
        self.title_rect_height = 39
        self.attr_rect_height = 26
        self.attr_rect_opacity = 0.9
        self.attr_plug_side_margin = 0
        self.plug_selection_range_mult = 1.1
        self.max_width = 260
        self.stack_offset = 20
        self.is_break = False
        self.is_start = False
        self.start_color = colors.ERROR
        self.is_proxy = False
        self.locked = False
        self.is_real = True
        self.attr_dots = [False, False, False]
        self.error_list = []
        self.error_item = None
        self.is_build_focus = False

        # local attributes
        self.user_attr_names = []
        self._attribute_draw_details = OrderedDict()
        self._attr_plug_graphics = {}
        self.exec_out_plug = NodeExecutionPlug(self.model, self.node_path, is_input=False, parent=self)
        self.exec_in_plug = NodeExecutionPlug(self.model, self.node_path, is_input=True, parent=self)
        self.is_hovered = False
        self.collapse_state = False
        self.collapse_arrows = []
        self.node_enabled = None
        self.node_instance = None

        self.model.build_idx_changed.connect(self.update_build_focus)
        self.model.executing_changed.connect(self.update_build_focus)
        # draw node
        self.update_from_model()

        # Setup groups
        # In
        self.in_anim_group = QtCore.QParallelAnimationGroup()
        self.in_anim_group.finished.connect(self.finished_anim)
        # Out
        self.out_anim_group = QtCore.QParallelAnimationGroup()
        self.out_anim_group.finished.connect(self.finished_anim)

    def _setup_anim_properties(self):
        # Position anim property
        self.pos_anim = QtCore.QPropertyAnimation(self, b"pos", self)
        # Set graphics effect
        effect = QtWidgets.QGraphicsOpacityEffect(self)
        effect.setOpacity(1)
        self.setGraphicsEffect(effect)
        # Opacity anim property
        self.opacity_anim = QtCore.QPropertyAnimation(effect, b"opacity",
                                                      effect)
        # Lower power caching
        self.setCacheMode(QtWidgets.QGraphicsItem.ItemCoordinateCache)

    def setup_in_anim(self):
        self._setup_anim_properties()
        self.in_anim_group.addAnimation(self.pos_anim)
        self.in_anim_group.addAnimation(self.opacity_anim)

    def setup_out_anim(self):
        self._setup_anim_properties()
        self.out_anim_group.addAnimation(self.pos_anim)
        self.out_anim_group.addAnimation(self.opacity_anim)

    def finished_anim(self):
        self.setGraphicsEffect(None)
        self.in_anim_group.clear()
        self.out_anim_group.clear()
        self.setCacheMode(QtWidgets.QGraphicsItem.DeviceCoordinateCache)
        self.view.update()
        self.view._animating.remove(self)

    def get_is_animating(self):
        i = self.in_anim_group.State.Running == self.in_anim_group.state()
        o = self.out_anim_group.State.Running == self.out_anim_group.state()
        return i or o

    def anim_into_place(self, end_pos):
        if self.get_is_animating():
            return
        if end_pos == self.pos():
            return
        self.view._animating.append(self)
        if self.view.do_animations:
            self.setup_in_anim()
        else:
            self.setPos(end_pos)
            self.in_anim_group.finished.emit()
            return

        self.opacity_anim.setStartValue(0)
        self.opacity_anim.setEndValue(1)
        self.opacity_anim.setDuration(80)

        curve = QtCore.QEasingCurve(QtCore.QEasingCurve.OutBack)
        curve.setAmplitude(.8)
        self.pos_anim.setEasingCurve(curve)

        self.pos_anim.setDuration(100)
        self.pos_anim.setEndValue(end_pos)
        self.in_anim_group.start()

    def anim_out(self):
        if self.get_is_animating():
            return
        self.view._animating.append(self)
        if self.view.do_animations:
            self.setup_out_anim()
        else:
            self.out_anim_group.finished.emit()
            return
        self.setCacheMode(QtWidgets.QGraphicsItem.ItemCoordinateCache)
        self.opacity_anim.setStartValue(1)
        self.opacity_anim.setEndValue(0)
        self.opacity_anim.setDuration(80)

        self.pos_anim.setDuration(80)
        self.pos_anim.setEasingCurve(QtCore.QEasingCurve.Linear)
        x_move = self.stack_offset * -1 * .5
        if not self.parentItem() or not self.parentItem().parentItem():
            y_move = 0.0
        else:
            y_move = (self.parentItem().boundingRect().height() * -1.0) * .5
        self.pos_anim.setEndValue(QtCore.QPointF(x_move, y_move))
        self.out_anim_group.start()

    def update_color(self):
        layers = self.model.get_layers_with_opinion(self.node_path)
        n_colors = []
        if layers:
            for l_path in layers:
                n_colors += [self.model.get_node_color(self.node_path, l_path)]
        elif not self.model.node_is_implied(self.node_path):
            disp_layer = self.model.comp_layer
            n_colors = [self.model.get_node_color(self.node_path, disp_layer)]
        if n_colors:
            self.colors = [QtGui.QColor(color_code) for color_code in n_colors]
            self.colors.reverse()
        else:
            self.colors = [QtGui.QColor(QtCore.Qt.darkGray)]
        if not self.is_real:
            self.color_alpha = 0.35
        else:
            self.color_alpha = 1.0
        self.dim_factor = 100
        if self.model.node_is_instance_child(self.node_path,
                                             self.model.comp_layer):
            self.dim_factor = 110

    def update_fonts(self):
        self.title_font.setStrikeOut(not self.node_enabled)
        self.node_instance = self.model.get_node_instance(self.node_path,
                                                          self.model.comp_layer)
        self.title_font.setItalic(bool(self.node_instance))

    @property
    def title_bounding_rect(self):
        """Returns the bounding rect of the title bar.

        :returns: bounding rect of the title bar
        :rtype: QtCore.QRectF
        """
        return QtCore.QRectF(0.0, 0.0, self.max_width, self.title_rect_height)

    @property
    def stack_bounding_rect(self):
        """Returns the bounding rect of this node and all it's descendants.

        :returns: bounding rect of this node and all it's descendants
        :rtype: QtCore.QRectF
        """
        out_rect = self.get_selection_rect()
        for child in self.childItems():
            if isinstance(child, NodeGraphicsItem):
                child_rect = child.stack_bounding_rect
                out_rect.setHeight(child_rect.height() + out_rect.height())
        return out_rect

    @property
    def screen_pos(self):
        node_pos = self.scenePos()
        view_pos = self.view.mapFromScene(node_pos)
        viewport_pos = self.view.viewport().mapToParent(view_pos)
        global_pos = self.view.mapToGlobal(QtCore.QPoint(0, 0))
        x = global_pos.x() + viewport_pos.x()
        y = global_pos.y() + viewport_pos.y()
        return QtCore.QPointF(x, y)

    def itemChange(self, change, value):
        """Override of QtWidgets.QGraphicsItem itemChange."""
        # keep connections drawing to node as it moves
        if change is self.ItemScenePositionHasChanged:
            graphics = self.view.get_node_connection_graphics(self.node_path)
            for connection in graphics:
                connection.rebuild_line()
        # TODO: Take into account the positions of every selected node and snap them all to a grid as soon as
        #  the user preses shift. This will avoid the weird wavy snapping effect we have right now
        if change == self.ItemPositionChange and self.scene():
            ml = QtWidgets.QApplication.mouseButtons() == QtCore.Qt.LeftButton
            shift = QtWidgets.QApplication.keyboardModifiers() == QtCore.Qt.ShiftModifier
            force_snap = self.view.alignment_actions.snap_action.isChecked()
            if (ml & shift) or force_snap and not self.get_is_animating():
                value = self.closest_grid_point(value)
                return value

        return super(NodeGraphicsItem, self).itemChange(change, value)

    def boundingRect(self):
        """Override of QtWidgets.QGraphicsItem boundingRect. If this rectangle does not encompass the entire
        drawn item, artifacting will happen.
        """
        return self.get_selection_rect()

    def paint(self, painter, option, widget):
        """Override of QtWidgets.QGraphicsItem paint. Handles all visuals of the Node. Split up into 3
        functions for organization.
        """
        lod = QtWidgets.QStyleOptionGraphicsItem.levelOfDetailFromTransform(painter.worldTransform())
        if lod > MIN_LOD:
            painter.setRenderHints(QtGui.QPainter.Antialiasing |
                                   QtGui.QPainter.TextAntialiasing |
                                   QtGui.QPainter.SmoothPixmapTransform)
        else:
            painter.setRenderHints(QtGui.QPainter.Antialiasing |
                                   QtGui.QPainter.TextAntialiasing |
                                   QtGui.QPainter.SmoothPixmapTransform, False)
        self.draw_title(painter, lod)
        self.draw_attributes(painter, lod)
        self.draw_border(painter, lod)

    def closest_grid_point(self, position):
        snapped_pos = self.model.snap_pos_to_grid((position.x(), position.y()))
        return QtCore.QPointF(*snapped_pos)

    def draw_border(self, painter, lod=1.):
        """Draws border, called exclusively by paint.

        :param painter: painter from paint.
        :type painter: QtGui.QPainter
        """

        if self.model.is_selected(self.node_path):
            color = colors.SELECTED
        elif self.is_build_focus:
            color = QtCore.Qt.red
        elif self.is_hovered:
            color = QtCore.Qt.white
        else:
            painter.setPen(QtCore.Qt.NoPen)
            return
        if self.is_proxy:
            pen = QtGui.QPen(color, 1, QtCore.Qt.PenStyle.DashLine)
        else:
            pen = QtGui.QPen(color)
        if self.locked:
            c = QtGui.QColor(self.colors[-1])
            c.setAlphaF(.3)
            b = QtGui.QBrush(c)
            painter.setBrush(b)
            painter.setPen(QtCore.Qt.NoPen)
        else:
            painter.setPen(pen)
            painter.setBrush(QtCore.Qt.NoBrush)
        rect = QtCore.QRectF(self.get_selection_rect().x() + 1,
                             self.get_selection_rect().y() + 1,
                             self.get_selection_rect().width() - 2,
                             self.get_selection_rect().height() - 2)
        painter.drawRoundedRect(rect, self.ROUND_X, self.ROUND_Y)

    def draw_title(self, painter, lod=1.):
        """Draw title of the node. Called exclusively in paint.

        :param painter: painter from paint.
        :type painter: QtGui.QPainter
        """
        # draw bg
        painter.setPen(QtCore.Qt.NoPen)
        bg = painter.background()
        bgm = painter.backgroundMode()
        if self.error_item:
            self.scene().removeItem(self.error_item)
            self.error_item.deleteLater()
        self.error_item = None
        if self.is_real and not self.locked:
            painter.setBackgroundMode(QtCore.Qt.OpaqueMode)
        else:
            painter.setBackgroundMode(QtCore.Qt.TransparentMode)
            [c.setAlphaF(self.color_alpha) for c in self.colors]
        color_count = len(self.colors)
        color_band_width = 10
        for i in range(color_count):
            color = self.colors[i]
            if self.is_proxy:
                painter.setBackground(color.darker(self.dim_factor))
                brush = QtGui.QBrush(color.darker(self.dim_factor*2),
                                     QtCore.Qt.FDiagPattern)
            else:
                brush = QtGui.QBrush(color.darker(self.dim_factor))
            if self.locked:
                c = color.darker(self.dim_factor)
                c.setAlphaF(.5)
                painter.setBackground(c)
                brush = QtGui.QBrush(c.darker(self.dim_factor * 2),
                                     QtCore.Qt.Dense1Pattern)
            painter.setBrush(brush)
            # Top Opinion
            if i+1 == color_count:
                remaining_width = self.max_width - (i*color_band_width)
                rect = QtCore.QRectF(0, 0, remaining_width,
                                     self.title_rect_height)
            # Lower Opinions
            else:
                x_pos = self.max_width - (i+1)*color_band_width
                rect = QtCore.QRectF(x_pos, 0, color_band_width,
                                     self.title_rect_height)
            painter.drawRoundedRect(rect, self.ROUND_X, self.ROUND_Y)
        painter.setBackground(bg)
        painter.setBackgroundMode(bgm)
        # draw exec plugs
        exec_attr = nxt_node.INTERNAL_ATTRS.EXECUTE_IN
        exec_in_pos = self.get_attr_in_pos(exec_attr, scene=False)
        out_pos = self.get_attr_out_pos(exec_attr, scene=False)
        self.exec_out_plug.setPos(out_pos)
        self.exec_in_plug.setPos(exec_in_pos)
        if lod > MIN_LOD:
            # draw attr dots
            offset = -6
            for fill in self.attr_dots:
                painter.setBrush(QtCore.Qt.white)
                if fill:
                    painter.setBrush(QtCore.Qt.white)
                else:
                    painter.setBrush(QtCore.Qt.NoBrush)
                dots_color = QtGui.QColor(QtCore.Qt.white).darker(self.dim_factor)
                painter.setPen(QtGui.QPen(dots_color, 0.5))
                dot_x = self.max_width - 15
                dot_y = (self.title_rect_height / 2) + offset
                painter.drawEllipse(QtCore.QPointF(dot_x, dot_y), 2, 2)
                offset += 6

        # draw title

        painter.setFont(self.title_font)
        title_str = nxt_path.node_name_from_node_path(self.node_path)
        font_metrics = QtGui.QFontMetrics(self.title_font)
        width = self.max_width - 40
        if self.error_list:
            width -= 20
        if lod > MIN_LOD:
            painter.setPen(
                QtGui.QColor(QtCore.Qt.white).darker(self.dim_factor))
            if not self.node_enabled:
                painter.setPen(QtGui.QColor(QtCore.Qt.white).darker(150))
            title = font_metrics.elidedText(title_str,
                                            QtCore.Qt.ElideRight, width)
            painter.drawText(15, 0, self.max_width - 15, self.title_rect_height,
                             QtCore.Qt.AlignVCenter, title)
        else:
            painter.setBrush(QtGui.QColor(QtCore.Qt.white).darker(self.dim_factor))
            if not self.node_enabled:
                painter.setBrush(QtGui.QColor(QtCore.Qt.white).darker(150))
            proxy_rect = font_metrics.boundingRect(title_str)
            r_width = proxy_rect.width() * .8
            height = proxy_rect.height()
            painter.drawRect(15, height * .8,
                             min(r_width, width), height * .2)

        if lod > MIN_LOD:
            # draw error
            if self.error_list:
                pos = QtCore.QPointF(self.max_width-45, self.title_rect_height/4)
                font_size = user_prefs.get(USER_PREF.FONT_SIZE, 16)
                error_item = ErrorItem(font=QtGui.QFont('Roboto',
                                                        font_size, 75),
                                       pos=pos, text='!')
                error_item.setParentItem(self)
                error_item.setZValue(50)
                self.error_item = error_item

        # draw collapse state arrow
        for arrow in self.collapse_arrows:
            self.scene().removeItem(arrow)
        if lod > MIN_LOD:
            self.collapse_arrows = []
            # TODO calculation needed arrows should be done outside drawing

            if self.collapse_state:
                des_colors = self.model.get_descendant_colors(self.node_path)
                filled = self.model.has_children(self.node_path)
                if not filled:
                    des_colors = [QtCore.Qt.white]
                elif not des_colors:
                    disp = self.model.comp_layer
                    des_colors = [self.model.get_node_color(self.node_path, disp)]
                i = 0
                num = len(des_colors)
                for c in des_colors:
                    arrow = CollapseArrow(self, filled=filled, color=c)
                    arrow_width = arrow.width * 1.1
                    center_offset = (arrow_width * (num * .5) - arrow_width * .5)
                    cur_offset = (i * arrow_width)
                    pos = ((self.max_width * .5) + center_offset - cur_offset)
                    arrow.setPos(pos, self.boundingRect().height())
                    self.collapse_arrows += [arrow]
                    i += 1

    def draw_attributes(self, painter, lod=1.):
        """Draw attributes for this node. Called exclusively by paint.

        :param painter: painter from paint.
        :type painter: QtGui.QPainter
        """
        for attr_name in self.user_attr_names:
            # draw bg rect
            attr_details = self._attribute_draw_details[attr_name]
            painter.setBrush(attr_details['bg_color'])
            painter.setPen(QtCore.Qt.NoPen)
            painter.drawRect(attr_details['bg_rect'])

            # draw connection points
            target_color = attr_details['plug_color']

            self._attr_plug_graphics.setdefault(attr_name, {})
            attr_plug_graphics = self._attr_plug_graphics[attr_name]
            current_in_plug = attr_plug_graphics.get('in_plug')
            if lod > MIN_LOD:
                in_pos = self.get_attr_in_pos(attr_name, scene=False)
                if current_in_plug:
                    current_in_plug.show()
                    current_in_plug.color = target_color
                    current_in_plug.update()
                else:
                    current_in_plug = NodeGraphicsPlug(radius=self.ATTR_PLUG_RADIUS,
                                                       color=target_color,
                                                       attr_name_represented=attr_name,
                                                       is_input=True,
                                                       parent=self)
                    attr_plug_graphics['in_plug'] = current_in_plug
                current_in_plug.setPos(in_pos)
            elif current_in_plug:
                current_in_plug.hide()

            current_out_plug = attr_plug_graphics.get('out_plug')
            if lod > MIN_LOD:
                out_pos = self.get_attr_out_pos(attr_name, scene=False)
                if current_out_plug:
                    current_out_plug.show()
                    current_out_plug.color = target_color
                    current_out_plug.update()
                else:
                    current_out_plug = NodeGraphicsPlug(radius=self.ATTR_PLUG_RADIUS,
                                                        color=target_color,
                                                        attr_name_represented=attr_name,
                                                        is_input=False,
                                                        parent=self)
                    attr_plug_graphics['out_plug'] = current_out_plug
                current_out_plug.setPos(out_pos)
            elif current_out_plug:
                current_out_plug.hide()

            # draw attr_name
            rect = attr_details['bg_rect']
            painter.setFont(attr_details['title_font'])
            font_metrics = QtGui.QFontMetrics(self.attr_font)
            title = font_metrics.elidedText(attr_name, QtCore.Qt.ElideRight,
                                            self.max_width - 20)
            if lod > MIN_LOD:
                painter.setPen(attr_details['title_color'])
                painter.drawText(rect.x() + 10, rect.y() - 1, rect.width(),
                                 rect.height(), QtCore.Qt.AlignVCenter, title)
            else:
                proxy_rect = font_metrics.boundingRect(title)
                height = proxy_rect.height()
                width = proxy_rect.width()
                painter.setBrush(attr_details['title_color'].darker(150))
                painter.drawRect(rect.x() + 10, rect.y() + height*.8,
                                 width, height*.2)

    def calculate_attribute_draw_details(self):
        """Calculate position of all known attr names. Details stored in
        self._attribute_draw_details. Public interface to get details is split
        into two functions, get_attr_in_pos and get_attr_out_pos.
        """
        self._attribute_draw_details = OrderedDict()
        index = -1
        comp_layer = self.model.comp_layer
        for attr_name in self.user_attr_names:
            index += 1
            draw_details = {}
            # sizes
            rect_y = self.title_rect_height + index * self.attr_rect_height
            rect_midpoint = rect_y + (self.attr_rect_height / 2)
            draw_details['text_height'] = rect_y
            draw_details['bg_rect'] = QtCore.QRectF(0, rect_y, self.max_width,
                                                    self.attr_rect_height)
            # positions
            draw_details['in_pos'] = QtCore.QPointF(self.attr_plug_side_margin,
                                                    rect_midpoint)
            out_x = self.max_width - self.attr_plug_side_margin
            draw_details['out_pos'] = QtCore.QPointF(out_x, rect_midpoint)
            # background color
            color = self.model.get_node_attr_color(self.node_path, attr_name,
                                                   comp_layer)
            bg_color = QtGui.QColor(color).darker(150).darker(self.dim_factor)
            bg_color.setAlphaF(self.attr_rect_opacity)
            draw_details['bg_color'] = bg_color
            # plug color
            type_ = self.model.get_node_attr_type(self.node_path, attr_name,
                                                  comp_layer)
            draw_details['plug_color'] = colors.ATTR_COLORS.get(type_,
                                                                QtCore.Qt.gray)
            # title color
            attr_is_instance = self.model.node_attr_is_instance(self.node_path,
                                                                attr_name,
                                                                comp_layer)
            dim_title = 150 if attr_is_instance else self.dim_factor
            white = QtGui.QColor(QtCore.Qt.white)
            draw_details['title_color'] = white.darker(dim_title)
            # font
            font = QtGui.QFont(self.attr_font.family(),
                               self.attr_font.pointSize(),
                               italic=attr_is_instance)
            draw_details['title_font'] = font

            self._attribute_draw_details[attr_name] = draw_details
        # Internal Attrs
        # Exec
        draw_details = {}
        in_pos = QtCore.QPointF(0, self.title_bounding_rect.height()*0.5)
        draw_details['in_pos'] = in_pos
        out_pos = QtCore.QPointF(self.max_width,
                                 self.title_bounding_rect.height()*0.5)
        draw_details['out_pos'] = out_pos
        draw_details['plug_color'] = QtGui.QColor(QtCore.Qt.white)
        exec_attr = nxt_node.INTERNAL_ATTRS.EXECUTE_IN
        self._attribute_draw_details[exec_attr] = draw_details
        # Inst
        draw_details = {}
        in_pos = QtCore.QPointF(0, (self.title_bounding_rect.height()/3) * 2)
        draw_details['in_pos'] = in_pos
        out_pos = QtCore.QPointF(self.max_width,
                                 (self.title_bounding_rect.height()/3) * 2)
        draw_details['out_pos'] = out_pos
        draw_details['plug_color'] = QtGui.QColor(QtCore.Qt.gray)
        inst_attr = nxt_node.INTERNAL_ATTRS.INSTANCE_PATH
        self._attribute_draw_details[inst_attr] = draw_details

    def get_selection_rect(self):
        """used by boundingRect and draw_border."""
        min_rect = self.get_title_and_attrs_rect()
        selection_scale_offset = 0.5
        out_x = min_rect.x() - selection_scale_offset
        out_y = min_rect.y() - selection_scale_offset
        out_width = min_rect.right() + (selection_scale_offset * 2)
        out_heighth = min_rect.bottom() + (selection_scale_offset * 2)
        return QtCore.QRectF(out_x, out_y, out_width, out_heighth)

    def get_title_and_attrs_rect(self):
        """Returns the rectangle that contains the title and all of the attrs

        :return: The rectangle that contains the title and all of the attrs.
        """
        attrs_height = len(self.user_attr_names) * self.attr_rect_height
        height = self.title_rect_height + attrs_height
        return QtCore.QRectF(0, 0, self.max_width, height)

    def hoverEnterEvent(self, event):
        """Override of QtWidgets.QGraphicsItem hoverEnterEvent."""
        self.is_hovered = True
        if self.view.view_actions.tooltip_action.isChecked():
            self.setToolTip(self.get_node_tool_tip())
        else:
            self.setToolTip('')
        super(NodeGraphicsItem, self).hoverEnterEvent(event)

    def hoverLeaveEvent(self, event):
        """Override of QtWidgets.QGraphicsItem hoverLeaveEvent."""
        self.is_hovered = False
        self.update()
        super(NodeGraphicsItem, self).hoverLeaveEvent(event)

    def update_build_focus(self):
        if not self.model.executing:
            self.is_build_focus = False
            self.update()
            return
        self.is_build_focus = self.node_path == self.model.get_build_focus()
        self.update()

    def update_from_model(self):
        """Sync the visible state of this node with the model."""
        # Shortened var names
        comp = self.model.comp_layer
        node_path = self.node_path
        # update position
        if nxt_path.get_parent_path(node_path) == nxt_path.WORLD:
            pos = self.model.get_node_pos(node_path)
            self.setPos(pos[0], pos[1])
        self.is_real = self.model.node_exists(node_path)
        self.is_proxy = self.model.get_node_is_proxy(node_path)
        self.locked = self.model.get_node_locked(node_path)
        self.collapse_state = self.model.get_node_collapse(self.node_path,
                                                           comp)
        self.node_enabled = self.model.get_node_enabled(self.node_path)
        self.update_color()
        self.update_build_focus()
        self._update_collapse()
        self.error_list = self.model.get_node_error(node_path, comp)
        self.update_fonts()
        # update exec plugs
        # update attributes
        # 0 = no attributes
        # 1 = only local attributes
        # 2 = local and instanced attributes
        # 3 = local, instanced, and inherited attributes
        attr_display_state = self.model.get_attr_display_state(node_path)
        if attr_display_state == 1:
            attr_names = self.model.get_node_local_attr_names(node_path,
                                                              comp)
        elif attr_display_state == 2:
            local_attr_names = self.model.get_node_local_attr_names(node_path,
                                                                    comp)
            inst_attrs = self.model.get_node_instanced_attr_names(node_path,
                                                                  comp)
            attr_names = set(local_attr_names + inst_attrs)
        elif attr_display_state == 3:
            attr_names = self.model.get_node_attr_names(node_path, comp)
        else:
            attr_names = []
        removed_attr_names = set(self.user_attr_names).difference(attr_names)
        self.user_attr_names = sorted(attr_names)

        for removed_name in removed_attr_names:
            if removed_name in self._attribute_draw_details:
                self._attribute_draw_details.pop(removed_name)
            if removed_name not in self._attr_plug_graphics:
                continue
            plug_grpahics = self._attr_plug_graphics.pop(removed_name)
            plug = plug_grpahics.get('in_plug')
            if plug:
                self.scene().removeItem(plug)
            plug = plug_grpahics.get('out_plug')
            if plug:
                self.scene().removeItem(plug)
        self.calculate_attribute_draw_details()
        for key, value in self._attribute_draw_details.items():
            in_plug = value.get('in_plug')
            if in_plug:
                in_plug.update()

            out_plug = value.get('out_plug')
            if out_plug:
                out_plug.update()

        self.is_break = self.model.get_is_node_breakpoint(node_path, comp)
        self.is_start = self.model.get_is_node_start(node_path, comp)
        sp_attr = INTERNAL_ATTRS.START_POINT
        self.start_color = self.model.get_node_attr_color(node_path, sp_attr,
                                                          comp)


        if attr_display_state == 1:
            self.attr_dots = [True, False, False]
        elif attr_display_state == 2:
            self.attr_dots = [True, True, False]
        elif attr_display_state == 3:
            self.attr_dots = [True, True, True]
        else:
            self.attr_dots = [False, False, False]

        self.update()

    def update_plugs(self):
        self.calculate_attribute_draw_details()
        self.update()

    def update_collapse(self):
        self._update_collapse()
        self.update()

    def _update_collapse(self):
        comp_layer = self.model.comp_layer
        self.collapse_state = self.model.get_node_collapse(self.node_path,
                                                           comp_layer)

    def get_node_tool_tip(self):
        if not self.is_real:
            return 'Implicit node'
        disp_layer = self.model.comp_layer
        node = disp_layer.lookup(self.node_path)
        node_data = self.model.stage.get_node_data(node, disp_layer,
                                                   include_inherit=True)
        path_key = INTERNAL_ATTRS.as_save_key(INTERNAL_ATTRS.NODE_PATH)
        node_data[path_key] = self.node_path
        column_1_width = 20
        column_2_width = 80
        title = self.elided_text(getattr(node, INTERNAL_ATTRS.NAME),
                                 column_1_width + column_2_width).upper()
        table = self.table_format(node_data, column_1_width, column_2_width)
        return title + '\n' + table

    @staticmethod
    def table_format(data, col_1_width, col_2_width):
        offset = col_1_width - 1
        lines = list()
        for k, v in data.items():
            col_1_out = '\n'.join(textwrap.wrap(str(k),
                                                offset)).ljust(col_1_width, '.')
            if isinstance(v, list):
                str_v = str([str(i) for i in v])
                col_2_out = ('\n ' + (' ' * offset)).join(textwrap.wrap(str_v,
                                                          col_2_width))
            else:
                col_2_out = ('\n ' +
                             (' ' * offset)).join(textwrap.wrap(str(v),
                                                                col_2_width))
            lines += [col_1_out + col_2_out]
        return '\n'.join(lines)

    @staticmethod
    def elided_text(text, width, char='.', repeat=3):
        return text if len(str(text)) < width else text[:width - repeat] + (char * repeat)

    def rename_node(self):
        """Presents a rename dialog pop-up over the node item. The new name is sent to the model."""
        # get dialog geometry
        scale_factor = self.view.transform().m11()
        scale = scale_factor if scale_factor > 0.75 else 0.75
        title_rect = self.title_bounding_rect.toRect()
        width = title_rect.width() * scale
        height = title_rect.height() * scale
        pos = self.screen_pos
        geometry = QtCore.QRect(pos.x(), pos.y(), width, height)

        # rename dialog
        name = nxt_path.node_name_from_node_path(self.node_path)
        dialog = NameEditDialog(name=name, geometry=geometry, parent=None)
        color_name = self.colors[-1].name()
        dialog.setStyleSheet('background-color: %s; color: white' % color_name)
        font = QtGui.QFont(self.title_font.family(), self.title_font.pointSizeF() * scale)
        dialog.line_edit.setFont(font)
        dialog.exec_()

        # rename node
        if dialog.result():
            new_name = dialog.value
            self.model.set_node_name(self.node_path, new_name, self.model.target_layer)

    def collapse_node(self, recursive=False):
        """Collapse the node hierarchy below this node"""
        is_collapsed = self.model.get_node_collapse(self.node_path)
        self.model.toggle_node_collapse(node_paths=[self.node_path],
                                        recursive_down=recursive,
                                        layer_path=LAYERS.TOP)

    def arrange_descendants(self):
        """Recursively arrange this node's descendants
        """
        self.stack_height = self.get_selection_rect().height()
        if self.model.get_node_collapse(self.node_path):
            return
        children_paths = self.model.get_children(self.node_path, ordered=True,
                                                 include_implied=True)
        prev_y = 0
        prev_child = None
        index = 1
        for child_path in children_paths:
            child = self.view.get_node_graphic(child_path)
            if not child:
                continue
            child.setParentItem(self)
            child.setFlag(QtWidgets.QGraphicsItem.ItemIsMovable, False)

            if prev_child:
                y = prev_child.stack_height
                child.stackBefore(prev_child)
            else:
                y = self.get_selection_rect().height()
            y += prev_y
            new_pos = QtCore.QPointF(self.stack_offset, y)
            child.anim_into_place(new_pos)
            prev_y = y
            child.arrange_descendants()
            prev_child = child
            index += 1
        if prev_child:
            self.stack_height = y + prev_child.stack_height

    def get_attr_in_pos(self, attr_name, scene=True):
        """Get the position of given attr's input graphic, optionally in
        screen coordinates.

        :param attr_name: Name of the attribute to get position for.
        :type attr_name: str
        :param scene: Whether or not to get the position in screen coordinates.
        :type scene: bool, defautls to True
        :return: Position of the attr in pin.
        """
        details = self._attribute_draw_details.get(attr_name)
        if not details:
            return
        pos = self._attribute_draw_details[attr_name]['in_pos']
        if not scene:
            return pos
        return self.mapToScene(pos)

    def get_attr_out_pos(self, attr_name, scene=True):
        """Get the position of given attr's output graphic, optionally in
        screen coordinates.

        :param attr_name: Name of the attribute to get position for.
        :type attr_name: str
        :param scene: Whether or not to get the position in screen coordinates.
        :type scene: bool, defaults to True

        :return: Position of the attr out pin.
        """
        details = self._attribute_draw_details.get(attr_name)
        if not details:
            return
        pos = self._attribute_draw_details[attr_name]['out_pos']
        if not scene:
            return pos
        return self.mapToScene(pos)


class NodeGraphicsPlug(QtWidgets.QGraphicsItem):

    """Graphics item for user attribute plugs on the NodeGraphicsItem."""

    def __init__(self, radius=3, hover_width=0.5, color=QtGui.QColor(255, 255, 255, 255),
                 attr_name_represented='', is_input=False, parent=None):
        super(NodeGraphicsPlug, self).__init__(parent=parent)
        self.setAcceptHoverEvents(True)
        # TODO benchmark this cache setting to see if it helps or hurts performance
        self.setCacheMode(QtWidgets.QGraphicsItem.DeviceCoordinateCache)

        self.radius = radius
        self.hover_width = hover_width
        self.color = QtGui.QColor(color)
        self.attr_name_represented = attr_name_represented
        self.is_input = is_input

        self.is_hovered = False

    def boundingRect(self):
        """Override of QtWidgets.QGraphicsItem boundingRect. If this rectangle does not encompass the entire
        drawn item, artifacting will happen.
        """
        offset = self.hover_width * 8
        return QtCore.QRect((self.radius + offset) * -1,
                            (self.radius + offset) * -1,
                            (self.radius + offset) * 2,
                            (self.radius + offset) * 2)

    def _apply_lod_to_painter(self, painter):
        lod = QtWidgets.QStyleOptionGraphicsItem.levelOfDetailFromTransform(
            painter.worldTransform())
        if lod > MIN_LOD:
            painter.setRenderHints(QtGui.QPainter.Antialiasing |
                                   QtGui.QPainter.TextAntialiasing |
                                   QtGui.QPainter.SmoothPixmapTransform)
        else:
            painter.setRenderHints(QtGui.QPainter.Antialiasing |
                                   QtGui.QPainter.TextAntialiasing |
                                   QtGui.QPainter.SmoothPixmapTransform, False)

    def paint(self, painter, option, widget):
        """Override of QtWidgets.QGraphicsItem paint. Handles all visuals of the Plug."""
        self._apply_lod_to_painter(painter)
        if self.is_hovered:
            painter.setPen(QtGui.QPen(QtCore.Qt.white, self.hover_width, QtCore.Qt.SolidLine, QtCore.Qt.RoundCap, QtCore.Qt.RoundJoin))
        else:
            painter.setPen(QtCore.Qt.NoPen)
        painter.setBrush(self.color)
        painter.drawEllipse(QtCore.QPointF(0, 0), self.radius, self.radius)

    def itemChange(self, change, value):
        """Override of QtWidgets.QGraphicsItem itemChange."""
        if change is QtWidgets.QGraphicsItem.ItemSceneChange:
            if value != self.scene():
                # the z value here is really only relevant in the context of a NodeGraphicsItem,
                # which is the only place these are ever made.
                self.setZValue(1)

                # this is the moment we've been removed from a scene.
                if not value:
                    del self
                    return

        return super(NodeGraphicsPlug, self).itemChange(change, value)

    def hoverEnterEvent(self, event):
        """Override of QtWidgets.QGraphicsItem hoverEnterEvent."""
        self.is_hovered = True
        self.update()
        super(NodeGraphicsPlug, self).hoverEnterEvent(event)

    def hoverLeaveEvent(self, event):
        """Override of QtWidgets.QGraphicsItem hoverLeaveEvent."""
        self.is_hovered = False
        self.update()
        super(NodeGraphicsPlug, self).hoverLeaveEvent(event)

    def mousePressEvent(self, event):
        """Override of QtWidgets.QGraphicsItem mousePressEvent."""
        # break attribute connections
        if event.modifiers() == QtCore.Qt.AltModifier:
            raise NotImplementedError("Alt to clear attribute is broken.")
        if self.attr_name_represented:
            node_path = self.parentItem().node_path
            path = nxt_path.make_attr_path(node_path, self.attr_name_represented)
        else:
            path = self.parentItem().node_path

        if self.is_input:
            self.parentItem().view.start_connection_draw(tgt_path=path)
        else:
            self.parentItem().view.start_connection_draw(src_path=path)
        event.accept()

        # Intentionally not sending this super call because we need to absorb this click when we are
        # dragging to make a connection.
        # super(NodeGraphicsPlug, self).mousePressEvent(event)

    def mouseReleaseEvent(self, event):
        """Override of QtWidgets.QGraphicsItem mouseReleaseEvent."""
        # parent_name = self.parentItem().node_path
        # print(parent_name + 'plug mouse press')
        super(NodeGraphicsPlug, self).mouseReleaseEvent(event)


class NodeExecutionPlug(NodeGraphicsPlug):
    """Node Graphics Plug for the plugs in the execution position.

    Handles drawing of exec plugs, as well as start, break, and skip points.
    """
    def __init__(self, model, node_path, is_input, parent=None):
        super(NodeExecutionPlug, self).__init__(
            radius=NodeGraphicsItem.EXEC_PLUG_RADIUS,
            color=QtCore.Qt.white,
            attr_name_represented=nxt_node.INTERNAL_ATTRS.EXECUTE_IN,
            is_input=is_input,
            parent=parent
        )
        self.node_path = node_path
        self.model = model

        self.is_root = None
        self.is_break = None
        self.is_start = None
        self.start_color = None
        self.is_skip = None
        self.model.starts_changed.connect(self._refresh_is_start)
        self.model.breaks_changed.connect(self._refresh_is_break)
        self.model.skips_changed.connect(self._refresh_is_skip)
        self._refresh_from_model()

    def _refresh_is_start(self):
        self.is_start = self.model.get_is_node_start(self.node_path, self.model.comp_layer)
        if self.is_start:
            self.start_color = self.model.get_node_attr_color(
                self.node_path, INTERNAL_ATTRS.START_POINT, self.model.comp_layer)
        self.update()

    def _refresh_is_break(self):
        self.is_break = self.model.get_is_node_breakpoint(self.node_path, self.model.comp_layer)
        self.update()

    def _refresh_is_skip(self):
        self.is_skip = self.model.is_node_skippoint(self.node_path, self.model.comp_layer.real_path)
        self.update()

    def _refresh_from_model(self):
        self.is_root = nxt_path.get_path_depth(self.node_path) == 1
        self._refresh_is_break()
        self._refresh_is_start()
        self._refresh_is_skip()

    def paint(self, painter, option, widget):
        """Override of QtWidgets.QGraphicsItem paint. Handles all visuals of the Plug."""
        special_input = any([self.is_break, self.is_start, self.is_skip])
        # If an output, or an non-special input.
        if (not self.is_input) or (not special_input):
            if self.is_root:
                self.radius = NodeGraphicsItem.EXEC_PLUG_RADIUS
                super(NodeExecutionPlug, self).paint(painter, option, widget)
            # For non-root non-specials, no drawing.
            return
        self._apply_lod_to_painter(painter)
        shape_border_pen = QtGui.QPen(colors.GRAPH_BG_COLOR, self.hover_width * 2)
        if self.is_start:  # A green triangle, with a border of layer color
            self.radius = 13
            painter.setBrush(QtGui.QColor(self.start_color))
            painter.setPen(shape_border_pen)
            painter.drawPolygon(self._buildTriangle(QtCore.QPointF(self.radius * -0.6, 0), self.radius))
            painter.setBrush(colors.START_COLOR)
            painter.setPen(QtCore.Qt.NoPen)
            painter.drawPolygon(self._buildTriangle(QtCore.QPointF(self.radius * -0.6, 0), self.radius * .7))
        elif self.is_skip:  # A short wide rect, a minus
            self.radius = 9
            painter.setBrush(colors.SKIP_COLOR)
            painter.setPen(shape_border_pen)
            skip_rect = QtCore.QRect(self.radius * -1, self.radius * -.6, self.radius * 2, self.radius)
            painter.drawRoundedRect(skip_rect, 4, 4)
        elif self.is_break:  # A red square
            self.radius = 8
            painter.setBrush(colors.BREAK_COLOR)
            painter.setPen(shape_border_pen)
            painter.drawRect(self.radius * -1, self.radius * -1, self.radius * 2, self.radius * 2)

    def _buildTriangle(self, offset, side_length):
        polygon = QtGui.QPolygonF()
        step_angle = 120
        for i in [0, 1, 2, 3]:
            step = step_angle * i
            x = side_length * 1.2 * math.cos(math.radians(step))
            y = side_length * 1.2 * math.sin(math.radians(step))
            polygon.append(QtCore.QPointF(x, y))
        polygon.translate(offset)
        return polygon


class CollapseArrow(QtWidgets.QGraphicsItem):

    """Graphics item for when NodeGraphicsItem stacks are collapsed."""

    def __init__(self, parent=None, filled=False, color=None):
        super(CollapseArrow, self).__init__(parent)
        self.height = 8
        self.width = 16
        self.filled = filled
        self.color = color or QtCore.Qt.white
        if sys.version_info[0] == 2:
            is_str = isinstance(self.color, basestring)
        else:
            is_str = isinstance(self.color, str)
        if is_str:
            self.color = QtGui.QColor(self.color)
        self.setZValue(30)

    def boundingRect(self):
        """Override of QtWidgets.QGraphicsItem boundingRect. If this rectangle
        does not encompass the entire drawn item, artifacting will happen.
        """
        return QtCore.QRect(0-(self.width*.5), 0, self.width, self.height)

    def paint(self, painter, option, widget):
        """Override of QtWidgets.QGraphicsItem paint."""
        painter.setRenderHints(QtGui.QPainter.Antialiasing |
                               QtGui.QPainter.TextAntialiasing |
                               QtGui.QPainter.SmoothPixmapTransform)
        if self.filled:
            brush = QtGui.QBrush(self.color)
            painter.setBrush(brush)
            if self.color.lightness() < 100:
                pen_color = QtGui.QColor(self.color.lighter(300))
            else:
                pen_color = QtGui.QColor(self.color.darker(300))
            pen = QtGui.QPen(pen_color)
            pen.setJoinStyle(QtCore.Qt.RoundJoin)
            painter.setPen(QtCore.Qt.NoPen)
        else:
            brush = QtGui.QBrush(QtCore.Qt.white, QtCore.Qt.Dense6Pattern)
            painter.setBrush(brush)
            pen = QtGui.QPen(QtCore.Qt.white)
            pen.setStyle(QtCore.Qt.DotLine)
            painter.setPen(pen)

        # draw triangle
        points = [
            QtCore.QPointF(0 - (self.width * .5), 0),
            QtCore.QPointF(self.width*.5, 0),
            QtCore.QPointF(0, self.height)
        ]
        painter.drawPolygon(points)
        # Draw outline
        if self.filled:
            painter.setBrush(QtCore.Qt.NoBrush)
            painter.setPen(pen)
            painter.drawLine(0 - (self.width * .5), 0,
                             0, self.height)
            painter.drawLine(0, self.height,
                             self.width * .5, 0)

    def itemChange(self, change, value):
        """Override of QtWidgets.QGraphicsItem itemChange."""
        if change is QtWidgets.QGraphicsItem.ItemSceneChange:
            if value != self.scene():
                # z value only relevant in the context of a NodeGraphicsItem,
                # which is the only place these are ever made.
                self.setZValue(1)

                # this is the moment we've been removed from a scene.
                if not value:
                    del self
                    return

        return super(CollapseArrow, self).itemChange(change, value)


class ErrorItem(QtWidgets.QGraphicsTextItem):
    def __init__(self, font, pos=QtCore.QPointF(), text='!', color=QtCore.Qt.red):
        super(ErrorItem, self).__init__()
        self.setPos(pos)
        self.text = text
        self.color = color
        self.setFont(font)

    def boundingRect(self):
        return QtCore.QRectF(-10, -10, 10, 10)

    def paint(self, painter, option, widget):
        painter.setRenderHints(QtGui.QPainter.Antialiasing |
                               QtGui.QPainter.TextAntialiasing)
        painter.setPen(QtCore.Qt.NoPen)
        painter.setBrush(colors.ERROR)
        painter.drawEllipse(0, 0, 20, 20)
        painter.setPen(QtCore.Qt.black)
        painter.setFont(self.font())
        painter.drawText(QtCore.QRectF(0.6, 0.1, 20, 20), QtCore.Qt.AlignCenter,
                         self.text)



# External
from Qt import QtWidgets
from Qt import QtGui
from Qt import QtCore


class LabelEdit(QtWidgets.QLabel):
    doubleClicked = QtCore.Signal()
    nameChangeRequested = QtCore.Signal(str)

    def __init__(self, *args, **kwargs):
        super(LabelEdit, self).__init__(*args, **kwargs)
        self.doubleClicked.connect(self.edit_text)
        self._read_only = False

    def setReadOnly(self, state):
        """Named this way to mimic Qt"""
        self._read_only = state

    def mouseDoubleClickEvent(self, event):
        if self._read_only:
            return
        if event.button() == QtCore.Qt.LeftButton:
            self.doubleClicked.emit()

    def edit_text(self):
        if self._read_only:
            return
        # get current name
        name = self.text()

        # get current geometry
        rect = self.rect()
        point = rect.topLeft()
        global_point = self.mapToGlobal(point)
        geometry = QtCore.QRect(global_point.x(), global_point.y() - 1, rect.width(), rect.height())

        # dialog
        dialog = NameEditDialog(name=name, geometry=geometry, parent=self)
        dialog.line_edit.setFont(self.font())
        dialog.exec_()

        # change name
        if dialog.result():
            self.nameChangeRequested.emit(dialog.value)


class NameEditDialog(QtWidgets.QDialog):
    """https://gist.github.com/justinfx/1951709"""

    def __init__(self, name='', geometry=None, parent=None):
        super(NameEditDialog, self).__init__(parent)
        self.setGeometry(geometry)
        self.setStyleSheet('background-color: #3E3E3E')
        self.setWindowFlags(QtCore.Qt.Window | QtCore.Qt.FramelessWindowHint | QtCore.Qt.Popup)

        # layout
        self.layout = QtWidgets.QHBoxLayout()
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self.layout)

        # line edit
        self.line_edit = QtWidgets.QLineEdit(name, parent=self)
        self.line_edit.setFocus()
        self.line_edit.selectAll()
        self.line_edit.setStyleSheet('background-color: #232323')
        self.layout.addWidget(self.line_edit)

    @property
    def value(self):
        return self.line_edit.text()

    def keyPressEvent(self, event):
        if event.key() == QtCore.Qt.Key_Escape:
            self.reject()
            self.done(0)
        elif event.key() == QtCore.Qt.Key_Return:
            self.accept()
            self.done(1)
        else:
            super(NameEditDialog, self).keyPressEvent(event)

    def closeEvent(self, event):
        self.accept()
        super(NameEditDialog, self).closeEvent(event)



# Builtin
import os
import logging
import sys

# External
from Qt import QtCore, QtWidgets, QtGui

# Internal
import nxt

logger = logging.getLogger('nxt.nxt_editor')

LOGGER_NAME = logger.name


class DIRECTIONS:
    UP = 'up'
    DOWN = 'down'
    LEFT = 'left'
    RIGHT = 'right'


class LoggingSignaler(QtCore.QObject):
    """Qt object used to emit logging messages. This object allows us to make
    thread safe visual loggers.
    """
    signal = QtCore.Signal(logging.LogRecord)


class StringSignaler(QtCore.QObject):
    """Qt object used to emit strings. This object allows us to use Qt
    signals in objects that themselves can't be a QObject.
    """
    signal = QtCore.Signal(str)


def make_resources(qrc_path=None, result_path=None):
    import PySide2
    pyside_dir = os.path.dirname(PySide2.__file__)
    full_pyside2rcc_path = os.path.join(pyside_dir, 'pyside2-rcc')
    full_rcc_path = os.path.join(pyside_dir, 'rcc')
    this_dir = os.path.dirname(os.path.realpath(__file__))
    if not qrc_path:
        qrc_path = os.path.join(this_dir, 'resources/resources.qrc')
    if not result_path:
        result_path = os.path.join(this_dir, 'qresources.py')
    msg = 'First launch nxt resource generation from {} to {}'
    logger.info(msg.format(qrc_path, result_path))
    import subprocess
    ver = ['-py2']
    if sys.version_info[0] == 3:
        ver += ['-py3']
    args = [qrc_path] + ver + ['-o', result_path]
    try:
        subprocess.check_call(['pyside2-rcc'] + args)
    except:
        pass
    else:
        return

    try:
        subprocess.check_call([full_pyside2rcc_path] + args)
    except:
        pass
    else:
        return
    try:
        subprocess.check_call([full_rcc_path, '-g', 'python', qrc_path,
                               '-o', result_path], cwd=pyside_dir)
    except:
        pass
    else:
        return
    try:
        subprocess.check_call(['rcc', '-g', 'python', qrc_path,
                               '-o', result_path], cwd=pyside_dir)
    except:
        raise Exception("Failed to generate UI resources using pyside2 rcc!"
                        " Reinstalling pyside2 may fix the problem. If you "
                        "know how to use rcc please build from: \"{}\" and "
                        "output to \"{}\"".format(qrc_path, result_path))
    else:
        return


try:
    from nxt_editor import qresources
except ImportError:
    make_resources()
    from nxt_editor import qresources


def _new_qapp():
    app = QtWidgets.QApplication.instance()
    create_new = False
    if not app:
        app = QtWidgets.QApplication
        app.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling, True)
        create_new = True
    os.environ["QT_AUTO_SCREEN_SCALE_FACTOR"] = "1"
    app.setEffectEnabled(QtCore.Qt.UI_AnimateCombo, False)
    if create_new:
        app = app(sys.argv)
    style_file = QtCore.QFile(':styles/styles/dark/dark.qss')
    style_file.open(QtCore.QFile.ReadOnly | QtCore.QFile.Text)
    stream = QtCore.QTextStream(style_file)
    app.setStyleSheet(stream.readAll())
    pixmap = QtGui.QPixmap(':icons/icons/nxt.svg')
    app.setWindowIcon(QtGui.QIcon(pixmap))
    return app


def launch_editor(paths=None, start_rpc=True):
    """Launch an instance of the editor. Will attach to existing QApp if found,
    otherwise will create and open one.
    """
    existing = QtWidgets.QApplication.instance()
    if existing:
        app = existing
    else:
        app = _new_qapp()
    instance = show_new_editor(paths, start_rpc)
    app.setActiveWindow(instance)
    if not existing:
        app.exec_()
    return instance


def show_new_editor(paths=None, start_rpc=True):
    path = None
    if paths is not None:
        path = paths[0]
        paths.pop(0)
    else:
        paths = []
    # Deferred import since main window relies on us
    from nxt_editor.main_window import MainWindow
    instance = MainWindow(filepath=path, start_rpc=start_rpc)
    for other_path in paths:
        instance.load_file(other_path)
    instance.show()
    return instance

